# coding=utf-8
# --------------------------------------------------------------------------------------------------------------------
# Licensed under the MIT LICENSE. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------------------------------
"""Simulation constants and default state variables."""
from bancor_research import Decimal

SIMULATION_TARGET_TVL: Decimal = Decimal("160000000")
SIMULATION_TARGET_TRADE_VOLUME: Decimal = Decimal("20000000")
SIMULATION_WHALE_THRESHOLD: Decimal = Decimal("1000000")
SIMULATION_TARGET_NUM_TRADES_PER_DAY: int = 200
SIMULATION_TARGET_NUM_DEPOSITS_PER_DAY: int = 200
SIMULATION_TARGET_NUM_WITHDRAWALS_PER_DAY: int = 200
SIMULATION_NUM_TRADERS = 2
SIMULATION_NUM_LPs = 2
SIMULATION_WHITELISTED_TOKENS = ["tkn", "bnt"]
SIMULATION_MAX_STEPS = 100
SIMULATION_OUTPUT_PATH: str = "data/simulation_output.csv"



from bancor_research.bancor_emulator.solidity import uint, uint8, uint16, uint32, uint256, hours, address, mapping, revert
from bancor_research.bancor_emulator.utils import contract, parse

from bancor_research.bancor_emulator.EnumerableSet import EnumerableSet
from bancor_research.bancor_emulator.Math import Math
from bancor_research.bancor_emulator.Constants import PPM_RESOLUTION
from bancor_research.bancor_emulator.Time import Time
from bancor_research.bancor_emulator.Vault import Vault as IVault
from bancor_research.bancor_emulator.RewardsMath import RewardsMath

# distribution types
FLAT_DISTRIBUTION = uint8(0);
EXP_DECAY_DISTRIBUTION = uint8(1);

class ProgramData:
    def __init__(self, x = None) -> None:
        self.startTime = parse(uint32, x, 'startTime');
        self.endTime = parse(uint32, x, 'endTime');
        self.halfLife = parse(uint32, x, 'halfLife');
        self.prevDistributionTimestamp = parse(uint32, x, 'prevDistributionTimestamp');
        self.poolToken = parse(address, x, 'poolToken');
        self.isPaused = parse(bool, x, 'isPaused');
        self.distributionType = parse(uint8, x, 'distributionType');
        self.totalRewards = parse(uint256, x, 'totalRewards');
        self.remainingRewards = parse(uint256, x, 'remainingRewards');

'''
 * @dev Auto-compounding Rewards contract
'''
class AutoCompoundingRewards(contract, Time):
    # the default number of programs to auto-process the rewards for
    DEFAULT_AUTO_PROCESS_REWARDS_COUNT = uint8(3);

    # the minimum time elapsed before the rewards of a program can be auto-processed
    AUTO_PROCESS_REWARDS_MIN_TIME_DELTA = uint16(1) * hours;

    # the factor used to calculate the maximum number of programs to attempt to auto-process in a single attempt
    AUTO_PROCESS_MAX_PROGRAMS_FACTOR = uint8(2);

    # if a program is attempting to burn a total supply percentage equal or higher to this number,
    # the program will terminate
    SUPPLY_BURN_TERMINATION_THRESHOLD_PPM = uint32(500000);

    '''
     * @dev a "virtual" constructor that is only used to set immutable state variables
    '''
    def __init__(self,
        initNetwork,
        initNetworkSettings,
        initBNT,
        initBNTPool,
        initExternalRewardsVault
    ) -> None:
        contract.__init__(self)
        Time.__init__(self)

        self._network = initNetwork;
        self._networkSettings = initNetworkSettings;
        self._bnt = initBNT;
        self._bntPool = initBNTPool;
        self._bntPoolToken = initBNTPool.poolToken();
        self._externalRewardsVault = initExternalRewardsVault;

    '''
     * @dev fully initializes the contract and its parents
    '''
    def initialize(self) -> None:
        self.__AutoCompoundingRewards_init();

    # solhint-disable func-name-mixedcase

    '''
     * @dev initializes the contract and its parents
    '''
    def __AutoCompoundingRewards_init(self) -> None:
        self.__AutoCompoundingRewards_init_unchained();

    '''
     * @dev performs contract-specific initialization
    '''
    def __AutoCompoundingRewards_init_unchained(self) -> None:
        # a mapping between pools and programs
        self._programs = mapping(lambda: ProgramData());

        # a set of all pools that have a rewards program associated with them
        self._pools = EnumerableSet();

        # the number of programs to auto-process the rewards for
        self._autoProcessRewardsCount = uint256();

        # the index of the next program to auto-process the rewards for
        self._autoProcessRewardsIndex = uint256();

        self._setAutoProcessRewardsCount(self.DEFAULT_AUTO_PROCESS_REWARDS_COUNT);

    # solhint-enable func-name-mixedcase

    '''
     * @inheritdoc Upgradeable
    '''
    def version(self) -> (int):
        return 1;

    '''
     * @inheritdoc IAutoCompoundingRewards
    '''
    def program(self, pool) -> (ProgramData):
        return self._programs[pool];

    '''
     * @inheritdoc IAutoCompoundingRewards
    '''
    def programs(self) -> (list):
        return self._pools._programs();

    '''
     * @inheritdoc IAutoCompoundingRewards
    '''
    def pools(self) -> (list):
        return self._pools.values();

    '''
     * @inheritdoc IAutoCompoundingRewards
    '''
    def autoProcessRewardsCount(self) -> (uint):
        return self._autoProcessRewardsCount.clone();

    '''
     * @dev sets the number of programs to auto-process the rewards for
     *
     * requirements:
     *
     * - the caller must be the admin of the contract
    '''
    def setAutoProcessRewardsCount(self, newCount) -> None:
        self._setAutoProcessRewardsCount(newCount);

    '''
     * @inheritdoc IAutoCompoundingRewards
    '''
    def isProgramActive(self, pool) -> (bool):
        p = self._programs[pool];

        if (not self._programExists(p)):
            return False;

        currTime = self._time();

        if (p.distributionType == EXP_DECAY_DISTRIBUTION):
            return p.startTime <= currTime;

        return p.startTime <= currTime and currTime <= p.endTime;

    '''
     * @inheritdoc IAutoCompoundingRewards
    '''
    def isProgramPaused(self, pool) -> (bool):
        return self._programs[pool].isPaused;

    '''
     * @inheritdoc IAutoCompoundingRewards
    '''
    def createFlatProgram(self,
        pool,
        totalRewards,
        startTime,
        endTime
    ) -> None:
        if (startTime >= endTime):
            revert("InvalidParam");

        self._createProgram(pool, totalRewards, FLAT_DISTRIBUTION, startTime, endTime, 0);

    '''
     * @inheritdoc IAutoCompoundingRewards
    '''
    def createExpDecayProgram(self,
        pool,
        totalRewards,
        startTime,
        halfLife
    ) -> None:
        if (halfLife == 0):
            revert("InvalidParam");

        self._createProgram(pool, totalRewards, EXP_DECAY_DISTRIBUTION, startTime, 0, halfLife);

    '''
     * @inheritdoc IAutoCompoundingRewards
    '''
    def terminateProgram(self, pool) -> None:
        self._terminateProgram(pool);

    '''
     * @inheritdoc IAutoCompoundingRewards
    '''
    def pauseProgram(self, pool, pause) -> None:
        p = self._programs[pool];

        if (not self._programExists(p)):
            revert("DoesNotExist");

        prevStatus = p.isPaused;
        if (prevStatus == pause):
            return;

        self._programs[pool].isPaused = pause;

    '''
     * @inheritdoc IAutoCompoundingRewards
    '''
    def autoProcessRewards(self) -> None:
        numOfPools = self._pools.length();
        index = self._autoProcessRewardsIndex.clone();
        count = self._autoProcessRewardsCount.clone();
        maxCount = Math.min(count * self.AUTO_PROCESS_MAX_PROGRAMS_FACTOR, numOfPools);

        for i in range(int(str(maxCount))):
            completed = self._processRewards((self._pools.at(index % numOfPools)), True);
            index += 1;
            if (completed):
                count -= 1;
                if (count == 0):
                    break;

        self._autoProcessRewardsIndex = index % numOfPools;

    '''
     * @inheritdoc IAutoCompoundingRewards
    '''
    def processRewards(self, pool) -> None:
        self._processRewards(pool, False);

    '''
     * @dev sets the number of programs to auto-process the rewards for
    '''
    def _setAutoProcessRewardsCount(self, newCount) -> None:
        prevCount = self._autoProcessRewardsCount;
        if (prevCount == newCount):
            return;

        self._autoProcessRewardsCount = newCount;

    '''
     * @dev processes the rewards of a given pool and returns true if the rewards processing was completed, and false
     * if it was skipped
    '''
    def _processRewards(self, pool, skipRecent) -> (bool):
        p = self._programs[pool];

        currTime = self._time();

        if (p.isPaused or currTime < p.startTime):
            return False;

        if (skipRecent and currTime < p.prevDistributionTimestamp + self.AUTO_PROCESS_REWARDS_MIN_TIME_DELTA):
            return False;

        tokenAmountToDistribute = self._tokenAmountToDistribute(p, currTime);
        if (tokenAmountToDistribute == 0):
            return True;

        poolTokenAmountToBurn = self._poolTokenAmountToBurn(pool, p, tokenAmountToDistribute);
        if (poolTokenAmountToBurn == 0):
            return True;

        # sanity check, if the amount to burn is equal or higher than the termination percentage
        # threshold, terminate the program
        if (
            poolTokenAmountToBurn * PPM_RESOLUTION >= p.poolToken.totalSupply() * self.SUPPLY_BURN_TERMINATION_THRESHOLD_PPM
        ):
            self._terminateProgram(pool);
            return False;

        rewardsVault = self._rewardsVault(pool);
        self._verifyFunds(poolTokenAmountToBurn, p.poolToken, rewardsVault);
        rewardsVault.burn(address(p.poolToken), poolTokenAmountToBurn);

        p.remainingRewards -= tokenAmountToDistribute;
        p.prevDistributionTimestamp = currTime;

        self._programs[pool] = p;

        return True;

    '''
     * @dev creates a rewards program for a given pool
    '''
    def _createProgram(self,
        pool,
        totalRewards,
        distributionType,
        startTime,
        endTime,
        halfLife
    ) -> None:
        if (self._programExists(self._programs[pool])):
            revert("AlreadyExists");

        poolToken = address(0);
        if (pool is (self._bnt)):
            poolToken = self._bntPoolToken;
        else:
            if (not self._networkSettings.isTokenWhitelisted(pool)):
                revert("NotWhitelisted");

            poolToken = self._network.collectionByPool(pool).poolToken(pool);

        if (startTime < self._time()):
            revert("InvalidParam");

        p = ProgramData({
            'startTime': startTime,
            'endTime': endTime,
            'halfLife': halfLife,
            'prevDistributionTimestamp': 0,
            'poolToken': poolToken,
            'isPaused': False,
            'distributionType': distributionType,
            'totalRewards': totalRewards,
            'remainingRewards': totalRewards
        });

        self._verifyFunds(self._poolTokenAmountToBurn(pool, p, totalRewards), poolToken, self._rewardsVault(pool));

        self._programs[pool] = p;

        assert(self._pools.add(address(pool)));

    '''
     * @dev terminates a rewards program
    '''
    def _terminateProgram(self, pool) -> None:
        p = self._programs[pool];

        if (not self._programExists(p)):
            revert("DoesNotExist");

        del self._programs[pool];

        assert(self._pools.remove(address(pool)));

    '''
     * @dev returns the amount of tokens to distribute
    '''
    def _tokenAmountToDistribute(self, p, currTime) -> (uint):
        prevTime = uint32(Math.max(p.prevDistributionTimestamp, p.startTime));

        if (p.distributionType == FLAT_DISTRIBUTION):
            currTimeElapsed = uint32(Math.min(currTime, p.endTime)) - p.startTime;
            prevTimeElapsed = uint32(Math.min(prevTime, p.endTime)) - p.startTime;
            return \
                RewardsMath.calcFlatRewards(p.totalRewards, currTimeElapsed - prevTimeElapsed, p.endTime - p.startTime);
        else:
            currTimeElapsed = currTime - p.startTime;
            prevTimeElapsed = prevTime - p.startTime;
            return \
                RewardsMath.calcExpDecayRewards(p.totalRewards, currTimeElapsed, p.halfLife) - \
                RewardsMath.calcExpDecayRewards(p.totalRewards, prevTimeElapsed, p.halfLife);

    '''
     * @dev returns the amount of pool tokens to burn
    '''
    def _poolTokenAmountToBurn(self,
        pool,
        p,
        tokenAmountToDistribute
    ) -> (uint):
        if (pool is (self._bnt)):
            return self._bntPool.poolTokenAmountToBurn(tokenAmountToDistribute);

        return \
            self._network.collectionByPool(pool).poolTokenAmountToBurn(
                pool,
                tokenAmountToDistribute,
                p.poolToken.balanceOf(address(self._externalRewardsVault))
            );

    '''
     * @dev returns whether or not a given program exists
    '''
    def _programExists(self, p) -> (bool):
        return address(p.poolToken) != address(0);

    '''
     * @dev returns the rewards vault for a given pool
    '''
    def _rewardsVault(self, pool) -> (IVault):
        return (self._bntPool) if pool is (self._bnt) else (self._externalRewardsVault);

    '''
     * @dev verifies that the rewards vault holds a sufficient amount of pool tokens
    '''
    def _verifyFunds(self,
        requiredAmount,
        poolToken,
        rewardsVault
    ) -> None:
        if (requiredAmount > poolToken.balanceOf(address(rewardsVault))):
            revert("InsufficientFunds");



from bancor_research.bancor_emulator.solidity import uint112, revert
from bancor_research.bancor_emulator.utils import library

from bancor_research.bancor_emulator.Fraction import Fraction256, Fraction112
from bancor_research.bancor_emulator.MathEx import MathEx

# solhint-disable-next-line func-visibility
def zeroFraction256() -> (Fraction256):
    return Fraction256({ 'n': 0, 'd': 1 });

# solhint-disable-next-line func-visibility
def zeroFraction112() -> (Fraction112):
    return Fraction112({ 'n': 0, 'd': 1 });

'''
 * @dev this library provides a set of fraction operations
'''
class FractionLibrary:
    isValid = None
    isPositive = None
    inverse = None
    toFraction112 = None
    fromFraction112 = None

'''
    * @dev returns whether a given fraction is valid
'''
def isValid(fraction) -> (bool):
    return fraction.d != 0;

'''
    * @dev returns whether a given fraction is positive
'''
def isPositive(fraction) -> (bool):
    return isValid(fraction) and fraction.n != 0;

'''
    * @dev returns the inverse of a given fraction
'''
def inverse(fraction):
    invFraction = type(fraction)({ 'n': fraction.d, 'd': fraction.n });

    if (not isValid(invFraction)):
        revert("InvalidFraction");

    return invFraction;

'''
    * @dev reduces a standard fraction to a 112-bit fraction
'''
def toFraction112(fraction: Fraction256) -> (Fraction112):
    truncatedFraction = MathEx.truncatedFraction(fraction, uint112.max);

    return Fraction112({ 'n': uint112(truncatedFraction.n), 'd': uint112(truncatedFraction.d) });

'''
    * @dev expands a 112-bit fraction to a standard fraction
'''
def fromFraction112(fraction: Fraction112) -> (Fraction256):
    return Fraction256({ 'n': fraction.n, 'd': fraction.d });

library(vars(), FractionLibrary)


# coding=utf-8
# --------------------------------------------------------------------------------------------------------------------
# Licensed under the MIT LICENSE. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------------------------------
"""Utility functions."""
from bancor_research.bancor_simulator.v3.spec.state import *
from bancor_research import Decimal, DataFrame
import json, pickle, cloudpickle


protocol_user = "protocol"


def check_if_program_enabled(start_time: int, end_time: int, timestamp: int):
    """
    Checks if the given rewards program is enabled.
    """
    return start_time <= timestamp <= end_time


def check_pool_shutdown(state: State, tkn_name: str) -> bool:
    """
    Checks that the bnt_min_trading_liquidity threshold has not been breached.
    """
    trading_enabled = get_is_trading_enabled(state, tkn_name)
    bnt_min_liquidity = get_bnt_min_liquidity(state, tkn_name)
    bnt_trading_liquidity = get_bnt_trading_liquidity(state, tkn_name)
    if bnt_trading_liquidity < bnt_min_liquidity and trading_enabled:
        return True
    return False


def check_is_bootstrap_reqs_met(
    state: State, tkn_name: str, bootstrap_liquidity: Decimal
) -> bool:
    """
    CHecks if the bootstrap requirements are met for a given tkn_name.
    """
    master_vault_balance = get_master_vault_balance(state, tkn_name)
    return master_vault_balance >= bootstrap_liquidity


def compute_user_total_holdings(state: State, user_name: str):
    """
    Returns the total number of holdings in this state.
    """
    holdings = []
    for tkn_name in state.users[user_name].wallet:
        amt = state.users[user_name].wallet[tkn_name].balance
        if len(tkn_name.replace("bn", "")) > 2:
            bntkn_rate = compute_bntkn_rate(state, tkn_name.replace("bn", ""))
            amt = amt / bntkn_rate
        holdings.append(amt)
    return sum(holdings)


def compute_rtkn_amt(state: State, tkn_name: str, ptkn_amt: Decimal) -> Decimal:
    """
    Returns the reserve token amount for a given pool token amount.
    """
    staked_balance = get_staked_balance(state, tkn_name)
    pool_token_supply = get_pooltoken_balance(state, tkn_name)
    return ptkn_amt * staked_balance / pool_token_supply


def compute_ptkn_amt(state: State, tkn_name: str, rtkn_amt: Decimal) -> Decimal:
    """
    Returns the pool token amount for a given reserve token amount.
    """
    staked_balance = get_staked_balance(state, tkn_name)
    pool_token_supply = get_pooltoken_balance(state, tkn_name)
    return rtkn_amt * pool_token_supply / staked_balance


def compute_bntkn_rate(state: State, tkn_name: str):
    """
    Computes the bntkn issuance rate for tkn deposits, based on the staking ledger and the current bntkn supply
    """
    pool_token_supply = get_pooltoken_balance(state, tkn_name)
    staked_balance = get_staked_balance(state, tkn_name)
    if pool_token_supply != staked_balance:
        return pool_token_supply / staked_balance
    return Decimal("1")


def compute_bntkn_amt(state: State, tkn_name: str, tkn_amt: Decimal) -> Decimal:
    """
    Returns the bntkn_amt for a given tkn_name, tkn_amt.
    """
    return compute_bntkn_rate(state, tkn_name) * tkn_amt


def compute_bnbnt_amt(state: State, bnt_amt: Decimal) -> Decimal:
    """
    Returns the bnbnt for a given bnt amt.
    """
    return state.bnbnt_rate * bnt_amt


def compute_changed_bnt_trading_liquidity(
    a: Decimal, b: Decimal, d: Decimal, e: Decimal, x: Decimal, direction: str
) -> Decimal:
    """
    Computes the changed state values according to the swap algorithm.
    """
    if direction == "tkn":
        return a * (b + d * x * (1 - e)) / (b + x)
    elif direction == "bnt":
        return (a * (a + x) + d * (1 - e) * (a * x + x**2)) / (a + d * x)


def compute_changed_tkn_trading_liquidity(
    a: Decimal, b: Decimal, d: Decimal, x: Decimal, direction: str
) -> Decimal:
    """
    Computes the changed state values according to the swap algorithm.
    """
    if direction == "tkn":
        return b + x
    elif direction == "bnt":
        return b * (a + d * x) / (a + x)


def compute_target_amt(
    a: Decimal, b: Decimal, d: Decimal, x: Decimal, direction: str
) -> Decimal:
    """
    Computes the changed state values according to the swap algorithm.
    """
    if direction == "tkn":
        return a * x * (1 - d) / (b + x)
    elif direction == "bnt":
        return b * x * (1 - d) / (a + x)


def compute_bootstrap_rate(state: State, tkn_name: str) -> Decimal:
    """
    Computes the bootstrap rate as bnt / tkn virtual balances.
    """
    return get_bnt_virtual_balance(state) / get_tkn_virtual_balance(state, tkn_name)


def compute_master_vault_tkn_tvl(
    master_vault_balance: Decimal, token_price: Decimal
) -> Decimal:
    """
    Computes the master vault tvl for a given tkn.
    """
    return master_vault_balance * token_price


def compute_max_tkn_deposit(
    master_vault_tvl: Decimal, target_tvl: Decimal, user_funds: Decimal
) -> Decimal:
    """
    Computes the maximum deposit amount which will not exceed the target tvl.
    """
    return max(target_tvl - master_vault_tvl, user_funds)


def format_json(val: Any, integer: bool = False, percentage: bool = False) -> Any:
    """
    Formats data when generating JSON test scenarios.
    """
    if integer:
        return str(val).replace("0E-18", "0", 10)
    elif percentage:
        return str(round(float(str(val)))).replace("0E-18", "0") + "%"
    else:
        if type(val) == dict:
            return val
        else:
            return str(val).replace("0E-18", "0")


def enable_trading(state: State, tkn_name: str) -> State:
    """
    Enables trading if the minimum bnt equivalent of tkn exists to justify bootstrapping.
    """

    assert not get_is_trading_enabled(state, tkn_name), "trading is already enabled"

    bootstrap_liquidity = get_tkn_bootstrap_liquidity(state, tkn_name)

    if check_is_bootstrap_reqs_met(state, tkn_name, bootstrap_liquidity):
        log = f"Bootstrap requirements met for {tkn_name}"
        state.logger.info(log)

        state.set_is_trading_enabled(tkn_name, True)

        bootstrap_rate = compute_bootstrap_rate(state, tkn_name)
        state.set_initial_rates(tkn_name, bootstrap_rate)

        bnt_bootstrap_liquidity = get_bnt_bootstrap_liquidity(state, tkn_name)
        state.set_bnt_trading_liquidity(tkn_name, bnt_bootstrap_liquidity)
        # changed due to contract implementation
        # should possibly use `set_bnt_funding_amt` instead
        state.increase_bnt_funding_amt(tkn_name, bnt_bootstrap_liquidity)

        tkn_bootstrap_liquidity = get_tkn_bootstrap_liquidity(state, tkn_name)
        state.set_tkn_trading_liquidity(tkn_name, tkn_bootstrap_liquidity)

        state = mint_protocol_bnt(state, bnt_bootstrap_liquidity)
    else:
        log = f"Bootstrap requirements *not* met for {tkn_name}"
        state.logger.info(log)
        raise Exception(log)

    return state


def mint_protocol_bnt(state: State, tkn_amt: Decimal) -> State:
    """
    Handles adjustments to the system resulting from the v3 minting BNT.
    """
    bnbnt_amt = compute_bnbnt_amt(state, tkn_amt)
    state.increase_master_vault_balance("bnt", tkn_amt)
    state.increase_pooltoken_balance("bnt", bnbnt_amt)
    state.increase_staked_balance("bnt", tkn_amt)
    state.increase_protocol_wallet_balance("bnt", bnbnt_amt)
    return state


def shutdown_pool(state: State, tkn_name: str) -> State:
    """
    Shutdown pool when the bnt_min_trading_liquidity threshold is breached.
    """

    bnt_trading_liquidity = get_bnt_trading_liquidity(state, tkn_name)
    bnt_renounced = min(bnt_trading_liquidity, get_bnt_funding_amt(state, tkn_name))
    bnbnt_renounced = compute_bnbnt_amt(state, bnt_renounced)

    # adjust balances
    state.decrease_staked_balance("bnt", bnt_renounced)
    state.decrease_master_vault_balance("bnt", bnt_trading_liquidity)
    state.decrease_pooltoken_balance("bnt", bnbnt_renounced)
    state.decrease_protocol_wallet_balance("bnt", bnbnt_renounced)
    state.decrease_bnt_funding_amt(tkn_name, bnt_renounced)

    # set balances
    state.set_is_trading_enabled(tkn_name, False)
    state.set_bnt_trading_liquidity(tkn_name, Decimal("0"))
    state.set_tkn_trading_liquidity(tkn_name, Decimal("0"))
    state.set_spot_rate(tkn_name, Decimal("0"))
    state.set_ema_rate(tkn_name, Decimal("0"))
    state.set_inv_spot_rate(tkn_name, Decimal("0"))
    state.set_inv_ema_rate(tkn_name, Decimal("0"))
    return state


def compute_pool_depth_adjustment(
    state: State, tkn_name: str, case: str = "none"
) -> Tuple[str, Decimal, Decimal]:
    """
    Computes the quantities of bnt and tkn to add to the pool trading liquidity during a deposit.
    """
    bnt_increase = tkn_increase = Decimal("0")
    is_trading_enabled = get_is_trading_enabled(state, tkn_name)
    is_price_stable = get_is_price_stable(state, tkn_name)
    bnt_remaining_funding = get_bnt_remaining_funding(state, tkn_name)
    if is_trading_enabled and is_price_stable:
        # changed due to contract implementation
        # should probably use `get_ema_rate` instead
        rate = get_updated_ema_rate(state, tkn_name)
        bnt_trading_liquidity = get_bnt_trading_liquidity(state, tkn_name)

        if bnt_remaining_funding > 0:
            tkn_excess = get_tkn_excess(state, tkn_name)
            tkn_excess_bnt_equivalence = get_tkn_excess_bnt_equivalence(
                state, tkn_name, rate
            )
            avg_tkn_trading_liquidity = get_avg_tkn_trading_liquidity(
                state, tkn_name, rate
            )

            if (
                avg_tkn_trading_liquidity <= tkn_excess
                and bnt_trading_liquidity <= bnt_remaining_funding
            ):
                case = "case1"
                bnt_increase = bnt_trading_liquidity
                tkn_increase = avg_tkn_trading_liquidity

            elif (
                avg_tkn_trading_liquidity <= tkn_excess
                and bnt_trading_liquidity > bnt_remaining_funding
                or avg_tkn_trading_liquidity > tkn_excess
                and tkn_excess_bnt_equivalence >= bnt_remaining_funding
            ):
                case = "case2"
                bnt_increase = bnt_remaining_funding
                tkn_increase = bnt_remaining_funding / rate

            elif (
                tkn_excess < avg_tkn_trading_liquidity
                and bnt_trading_liquidity <= bnt_remaining_funding
                or avg_tkn_trading_liquidity > tkn_excess
                and bnt_trading_liquidity
                > bnt_remaining_funding
                > tkn_excess_bnt_equivalence
            ):
                case = "case3"
                bnt_increase = tkn_excess_bnt_equivalence
                tkn_increase = tkn_excess

            else:
                raise ValueError(
                    "Something went wrong, pool adjustment case not found..."
                )

        # added due to contract implementation
        # was originally intended for admin only
        elif bnt_remaining_funding < 0:
            case = "case4"
            bnt_increase = bnt_remaining_funding
            tkn_increase = bnt_remaining_funding / rate

    return case, bnt_increase, tkn_increase


def vortex_collection(
    a: Decimal, b: Decimal, d: Decimal, e: Decimal, x: Decimal, direction: str
) -> Decimal:
    """
    Computes the changed state values according to the swap algorithm.
    """
    if direction == "tkn":
        return a * d * e * x / (b + x)
    elif direction == "bnt":
        return d * e * x * (a + x) / (a + d * x)


def swap_fee_collection(
    a: Decimal, b: Decimal, d: Decimal, e: Decimal, x: Decimal, direction: str
) -> Decimal:
    """
    Computes the swap fees according to the swap algorithm.
    """
    if direction == "tkn":
        return a * d * x * (1 - e) / (b + x)
    elif direction == "bnt":
        return b * d * x * (1 - e) / (a + x)


def handle_whitelisting_tokens(state):
    whitelisted_tokens = state.whitelisted_tokens
    bnt_min_liquidity = state.bnt_min_liquidity
    withdrawal_fee = state.withdrawal_fee
    network_fee = state.network_fee
    cooldown_time = state.cooldown_time

    for tkn_name in whitelisted_tokens:

        # Get tokens not yet initialized.
        if tkn_name not in state.tokens:

            decimals = whitelisted_tokens[tkn_name]["decimals"]
            trading_fee = whitelisted_tokens[tkn_name]["trading_fee"]
            bnt_funding_limit = whitelisted_tokens[tkn_name]["bnt_funding_limit"]
            ep_vault_balance = whitelisted_tokens[tkn_name]["ep_vault_balance"]

            # initialize tokens
            state.tokens[tkn_name] = Tokens(
                tkn_name=tkn_name,
                decimals=decimals,
                trading_fee=trading_fee,
                bnt_min_liquidity=bnt_min_liquidity,
                network_fee=network_fee,
                bnt_funding_limit=bnt_funding_limit,
                withdrawal_fee=withdrawal_fee,
                cooldown_time=cooldown_time,
                external_protection_vault=Token(balance=ep_vault_balance),
            )

            # initialize pooltoken
            pooltkn_name = get_pooltoken_name(tkn_name)
            state.tokens[pooltkn_name] = Tokens(
                tkn_name=pooltkn_name,
                trading_fee=trading_fee,
                bnt_min_liquidity=bnt_min_liquidity,
                network_fee=network_fee,
                bnt_funding_limit=bnt_funding_limit,
                withdrawal_fee=withdrawal_fee,
                cooldown_time=cooldown_time,
            )


def init_protocol(
    state: State,
    whitelisted_tokens: dict,
    usernames: List[str],
    cooldown_time: int,
    network_fee: Decimal,
    bnt_min_liquidity: Decimal,
    withdrawal_fee: Decimal,
) -> State:
    """
    Initialize user wallets upon system genesis.
    """

    for tkn_name in whitelisted_tokens:
        if tkn_name not in state.whitelisted_tokens:
            state.create_whitelisted_tkn(tkn_name)

    for tkn_name in whitelisted_tokens:

        # Get tokens not yet initialized.
        if tkn_name not in state.tokens:

            decimals = whitelisted_tokens[tkn_name]["decimals"]
            trading_fee = whitelisted_tokens[tkn_name]["trading_fee"]
            bnt_funding_limit = whitelisted_tokens[tkn_name]["bnt_funding_limit"]
            ep_vault_balance = whitelisted_tokens[tkn_name]["ep_vault_balance"]

            # initialize tokens
            state.tokens[tkn_name] = Tokens(
                tkn_name=tkn_name,
                decimals=decimals,
                trading_fee=trading_fee,
                bnt_min_liquidity=bnt_min_liquidity,
                network_fee=network_fee,
                bnt_funding_limit=bnt_funding_limit,
                withdrawal_fee=withdrawal_fee,
                cooldown_time=cooldown_time,
                external_protection_vault=Token(balance=ep_vault_balance),
            )

            # initialize pooltoken
            pooltkn_name = get_pooltoken_name(tkn_name)
            state.tokens[pooltkn_name] = Tokens(
                tkn_name=pooltkn_name,
                trading_fee=trading_fee,
                bnt_min_liquidity=bnt_min_liquidity,
                network_fee=network_fee,
                bnt_funding_limit=bnt_funding_limit,
                withdrawal_fee=withdrawal_fee,
                cooldown_time=cooldown_time,
            )

    for tkn_name in ["bnt", "bnbnt", "vbnt"]:

        state.tokens[tkn_name] = Tokens(
            tkn_name=tkn_name,
            network_fee=network_fee,
            withdrawal_fee=withdrawal_fee,
            cooldown_time=cooldown_time,
        )

    for usr in usernames:

        # Get users not yet initialized.
        if usr not in state.usernames:
            # initialize users.
            state.create_user(usr)

    state.whitelisted_tokens = whitelisted_tokens
    state.active_users = usernames
    return state


def handle_logging(
    tkn_name: str,
    tkn_amt: Decimal,
    action_name: str,
    user_name: str,
    transaction_id: int,
    state: State,
) -> State:
    """
    Logs the system state history after each action.
    """
    state.iter_transaction_id = transaction_id
    for tkn_name in state.whitelisted_tokens:
        # try:
        state_variables = {
            "timestamp": [state.timestamp],
            "latest_action": [action_name],
            "latest_amt": [tkn_amt],
            "latest_user_name": [user_name],
            "tkn_name": [tkn_name],
            "master_vault_tkn": [get_master_vault_balance(state, tkn_name)],
            "erc20contracts_bntkn": [get_pooltoken_balance(state, tkn_name)],
            "staked_tkn": [get_staked_balance(state, tkn_name)],
            "is_trading_enabled": [get_is_trading_enabled(state, tkn_name)],
            "bnt_trading_liquidity": [get_bnt_trading_liquidity(state, tkn_name)],
            "tkn_trading_liquidity": [get_tkn_trading_liquidity(state, tkn_name)],
            "trading_fee": [get_trading_fee(state, tkn_name)],
            "bnt_funding_limit": [get_bnt_funding_limit(state, tkn_name)],
            "bnt_remaining_funding": [get_bnt_remaining_funding(state, tkn_name)],
            "bnt_funding_amt": [get_bnt_funding_amt(state, tkn_name)],
            "external_protection_vault": [
                get_external_protection_vault_balance(state, tkn_name)
            ],
            "spot_rate": [get_spot_rate(state, tkn_name)],
            "ema_rate": [get_ema_rate(state, tkn_name)],
            "inv_spot_rate": [get_inv_spot_rate(state, tkn_name)],
            "inv_ema_rate": [get_inv_ema_rate(state, tkn_name)],
            "ema_last_updated": [state.tokens[tkn_name].ema_last_updated],
            "network_fee": [state.network_fee],
            "withdrawal_fee": [state.withdrawal_fee],
            "bnt_min_liquidity": [state.bnt_min_liquidity],
            "cooldown_time": [state.cooldown_time],
            "protocol_wallet_bnbnt": [get_protocol_wallet_balance(state, "bnt")],
            "vortex_bnt": [get_vortex_balance(state, "bnt")],
            "erc20contracts_bnbnt": [get_pooltoken_balance(state, "bnt")],
            "master_vault_bnt": [get_master_vault_balance(state, "bnt")],
            "staked_bnt": [get_staked_balance(state, "bnt")],
            "bnbnt_rate": [state.bnbnt_rate],
        }
        state.history.append(DataFrame(state_variables))
        # except KeyError:
        #     pass

    return state


def handle_vandalism_attack(state: State, tkn_name):
    """
    Checks if a vandalism attack has occured and adjusts the system state accordingly.
    """
    staked_tkn = get_staked_balance(state, tkn_name)
    bntkn = get_pooltoken_balance(state, tkn_name)
    if staked_tkn > 0 and bntkn == 0:
        state.set_staked_balance(tkn_name, Decimal("0"))
        shutdown_pool(state, tkn_name)
    return state


def handle_ema(state: State, tkn_name: str) -> State:
    """
    Handles the updating of the ema rate and the inverse ema rate, called before a swap is performed.
    """
    if state.tokens[tkn_name].ema_last_updated != state.tokens[tkn_name].timestamp:
        state.tokens[tkn_name].ema_last_updated = state.tokens[tkn_name].timestamp
        state.set_ema_rate(tkn_name, state.tokens[tkn_name].updated_ema_rate)
        state.set_inv_ema_rate(tkn_name, state.tokens[tkn_name].updated_inv_ema_rate)
    return state


def describe_rates(state: State, report={}) -> DataFrame:
    """
    Return a dataframe of the current system EMA & spot rates.
    """
    for tkn in state.whitelisted_tokens:
        report[tkn] = get_rate_report(state, tkn)
    return DataFrame(report).T.reset_index()


def describe(state: State, rates: bool = False, decimals=6) -> DataFrame:
    """
    Builds a dataframe of the current system/ledger state.
    """
    qdecimals = Decimal(10) ** -decimals
    if rates:
        return describe_rates(state, qdecimals=qdecimals)

    description = get_description(state, qdecimals)
    max_rows = max([len(description[key]) for key in description])

    # fill-in the description spacing for display purposes
    for col in description:
        while len(description[col]) < max_rows:
            description[col].append("")

    return DataFrame(description)


def build_json_operation(
    state: State,
    tkn_name: str,
    tkn_amt: Decimal,
    transaction_type: str,
    user_name: str,
    timestamp: int,
) -> dict:
    if "tkn" in state.autocompounding_reward_programs:
        program_wallet_bntkn = get_protocol_wallet_balance(state, tkn_name)
        tkn_remaining_rewards = get_autocompounding_remaining_rewards(state, tkn_name)
    else:
        program_wallet_bntkn = Decimal("0")
        tkn_remaining_rewards = Decimal("0")

    if "bnt" in state.autocompounding_reward_programs:
        bnt_remaining_rewards = get_autocompounding_remaining_rewards(state, "bnt")
    else:
        bnt_remaining_rewards = Decimal("0")

    if "tkn" in state.standard_reward_programs:
        er_vault_tkn = state.tokens[tkn_name].protocol_wallet_pooltokens.balance
    else:
        er_vault_tkn = Decimal("0")

    if "bnt" in state.autocompounding_reward_programs:
        er_vault_bnt = state.tokens["bnt"].protocol_wallet_pooltokens.balance
    else:
        er_vault_bnt = Decimal("0")

    if "User" not in state.users:
        print("User not found, therefore JSON export will not be created.")
        return {}

    json_operation = {
        "type": transaction_type,
        "userId": user_name,
        "elapsed": timestamp,
        "amt": format_json(tkn_amt) if str(tkn_amt) != "NA" else tkn_amt,
        "expected": {
            "tknBalances": {
                "User": format_json(get_user_balance(state, tkn_name, user_name)),
                "masterVault": format_json(get_master_vault_balance(state, tkn_name)),
                "erVault": format_json(er_vault_tkn),
                "epVault": format_json(
                    get_external_protection_vault_balance(state, tkn_name)
                ),
            },
            "bntBalances": {
                "User": format_json(get_user_balance(state, "bnt", user_name)),
                "masterVault": format_json(get_master_vault_balance(state, "bnt")),
                "erVault": format_json(er_vault_bnt),
            },
            "bntknBalances": {
                "User": format_json(
                    get_user_balance(state, f"bn{tkn_name}", user_name)
                ),
                "TKNProgramWallet": format_json(program_wallet_bntkn),
            },
            "bnbntBalances": {
                "User": format_json(get_user_balance(state, "bnbnt", user_name)),
                "bntPool": format_json(get_protocol_wallet_balance(state, "bnt")),
            },
            "tknRewardsRemaining": format_json(tkn_remaining_rewards),
            "bntRewardsRemaining": format_json(bnt_remaining_rewards),
            "bntCurrentPoolFunding": format_json(get_bnt_funding_amt(state, tkn_name)),
            "tknStakedBalance": format_json(get_staked_balance(state, tkn_name)),
            "bntStakedBalance": format_json(get_staked_balance(state, "bnt")),
            "tknTradingLiquidity": format_json(
                get_tkn_trading_liquidity(state, tkn_name)
            ),
            "bntTradingLiquidity": format_json(
                get_bnt_trading_liquidity(state, tkn_name)
            ),
            "averageRateN": format_json(
                state.tokens[tkn_name].ema.numerator, integer=True
            ),
            "averageRateD": format_json(
                state.tokens[tkn_name].ema.denominator, integer=True
            ),
            "averageInvRateN": format_json(
                state.tokens[tkn_name].inv_ema.numerator, integer=True
            ),
            "averageInvRateD": format_json(
                state.tokens[tkn_name].inv_ema.denominator, integer=True
            ),
        },
    }
    return json_operation


def log_json_operation(state, transaction_type, user_name, amt, timestamp):
    """Logs the latest operation for json testing."""

    tkn_name = [tkn for tkn in state.whitelisted_tokens if tkn != "bnt"][0]
    json_operation = build_json_operation(
        state, tkn_name, amt, transaction_type, user_name, timestamp
    )

    if transaction_type == "createAutocompoundingRewardProgramTKN":
        json_operation["amt"] = {}
        json_operation["amt"]["tknRewardsTotalamt"] = format_json(
            state.autocompounding_reward_programs["tkn"].total_rewards
        )
        json_operation["amt"]["tknRewardsStartTime"] = format_json(
            state.autocompounding_reward_programs["tkn"].start_time, integer=True
        )
        json_operation["amt"]["tknRewardsType"] = str(
            state.autocompounding_reward_programs["tkn"].distribution_type
        )
        json_operation["amt"]["tknHalfLife"] = format_json(
            state.autocompounding_reward_programs["tkn"].half_life, integer=True
        )

    if transaction_type == "createAutocompoundingRewardProgramBNT":
        state.json_export["bntRewardsTotalamt"] = format_json(
            state.autocompounding_reward_programs["bnt"].total_rewards
        )
        state.json_export["bntRewardsStartTime"] = format_json(
            state.autocompounding_reward_programs["bnt"].start_time, integer=True
        )
        state.json_export["bntRewardsType"] = str(
            state.autocompounding_reward_programs["bnt"].distribution_type
        )
        state.json_export["bntHalfLife"] = format_json(
            state.autocompounding_reward_programs["bnt"].half_life, integer=True
        )

    return json_operation


def validate_input(
    state: State, tkn_name: str, user_name: str, timestamp: int
) -> Tuple[State, str, str]:
    """
    Validates the input for all agent actions.
    """

    assert type(tkn_name) is str, "tkn_name must be of type string"
    assert type(user_name) is str, "tkn_name must be of type string"
    assert user_name != protocol_user, "user_name {} is reserved".format(protocol_user)

    tkn_name = tkn_name.lower()
    user_name = user_name if user_name else protocol_user

    if user_name not in state.users:
        state = state.create_user(user_name)

    if tkn_name not in state.users[user_name].wallet:
        state.users[user_name].wallet[tkn_name] = Token(balance=DEFAULT_ACCOUNT_BALANCE)

    pooltkn_name = get_pooltoken_name(tkn_name)
    if pooltkn_name not in state.users[user_name].wallet:
        state.users[user_name].wallet[pooltkn_name] = Token(
            balance=DEFAULT_ACCOUNT_BALANCE
        )

    if "vbnt" not in state.users[user_name].wallet:
        state.users[user_name].wallet["vbnt"] = Token(balance=DEFAULT_ACCOUNT_BALANCE)

    if "bnbnt" not in state.users[user_name].wallet:
        state.users[user_name].wallet["bnbnt"] = Token(balance=DEFAULT_ACCOUNT_BALANCE)

    state.tokens[tkn_name].timestamp = timestamp

    return state, tkn_name, user_name


def setup_json_simulation(
    state: State, json_data: dict, tkn_name: str, id_number: int
) -> State:
    """This method uploads a pre-formatted JSON file containing simulation modules to run and report on."""

    trading_fee = Decimal(json_data["tradingFee"].replace("%", "")) * Decimal(".01")

    if "tknRewardsamt" in json_data:
        try:
            state.standard_reward_programs[id_number].total_rewards = Decimal(
                json_data["tknRewardsamt"]
            )
            state.standard_reward_programs[id_number].end_time = int(
                json_data["tknRewardsDuration"]
            )
            state.standard_reward_programs["bnt"].total_rewards = Decimal(
                json_data["bntRewardsamt"]
            )
            state.standard_reward_programs["bnt"].end_time = int(
                json_data["bntRewardsDuration"]
            )
        except:
            pass

    # set params
    state.network_fee = Decimal(json_data["networkFee"].replace("%", "")) * Decimal(
        ".01"
    )

    state.withdrawal_fee = Decimal(
        json_data["withdrawalFee"].replace("%", "")
    ) * Decimal(".01")

    state.tokens[tkn_name].external_protection_vault = Decimal(
        json_data["epVaultBalance"]
    )

    state.decimals = Decimal(json_data["tknDecimals"])
    state.bnt_min_liquidity = Decimal(json_data["bntMinLiquidity"])
    state.bnt_funding_limit = Decimal(json_data["bntFundingLimit"])
    state.trading_fee = trading_fee

    for tkn_name in ["bnt", tkn_name]:
        state.tokens[tkn_name].trading_fee = trading_fee
        state.tokens[tkn_name].bnt_min_liquidity = state.bnt_min_liquidity
        state.tokens[tkn_name].bnt_funding_limit = state.bnt_funding_limit

    for user in json_data["users"]:
        user_name = user["id"]
        state.create_user(user_name)
        tknBalance = Decimal(user["tknBalance"])
        bntBalance = Decimal(user["bntBalance"])
        state.set_user_balance(user_name, "bnt", bntBalance)
        state.set_user_balance(user_name, "tkn", tknBalance)

    # set data to iterate
    state.json_data = json_data
    return state


def generate_emulator_expected_results(
    state: State,
    tkn_name: str,
    tkn_amt: Decimal,
    transaction_type: str,
    user_name: str,
    timestamp: int,
) -> dict:
    """
    Get the expected results for a given transaction.
    """
    json_operation = build_json_operation(
        state, tkn_name, tkn_amt, transaction_type, user_name, timestamp
    )
    return json_operation["expected"]


def load_json(path, **kwargs):
    """
    Loads json files for convenient simulation.
    """
    with open(path, "r") as f:
        return json.load(f, **kwargs)


def save_json(x, path, indent=True, **kwargs):
    """
    Saves json for convenience.
    """
    with open(path, "w") as f:
        if indent:
            json.dump(x, f, indent="\t", **kwargs)
        else:
            json.dump(x, f, **kwargs)
    print("Saved to", path)


def load_pickle(path):
    """
    Loads a pickled state from a file path.
    """
    print("Unpickling from", path)
    with open(path, "rb") as f:
        return pickle.load(f)


def save_pickle(x, path):
    """
    Saves a serialized state at file path.
    """
    print("Pickling to", path)
    with open(path, "wb") as f:
        return pickle.dump(x, f)


def load(file_path):
    """
    Loads saved file via cloudpickle.
    """
    with open(file_path, "rb") as f:
        return cloudpickle.load(f)


def export_test_scenarios(state: State, path: str = "test_scenarios.json"):
    """
    Exports the auto-generated json scenarios file to a given path.
    """
    save_json(state.json_export, path)



from bancor_research.bancor_emulator.solidity import uint256, uint112
from bancor_research.bancor_emulator.utils import parse

class Fraction:
    def __init__(self, cast, f) -> None:
        self.n = parse(cast, f, 'n')
        self.d = parse(cast, f, 'd')

    def __str__(self) -> str:
        return '{} / {}'.format(self.n, self.d)

class Fraction256(Fraction):
    def __init__(self, f) -> None:
        super().__init__(uint256, f)

class Fraction112(Fraction):
    def __init__(self, f) -> None:
        super().__init__(uint112, f)



from bancor_research.bancor_emulator.solidity import uint, uint8, uint16, uint32, uint64 ,uint96 ,uint128 ,uint224 ,uint256, require

'''
 * @dev Wrappers over Solidity's uintXX/intXX casting operators with added overflow
 * checks.
 *
 * Downcasting from uint256/int256 in Solidity does not revert on overflow. This can
 * easily result in undesired exploitation or bugs, since developers usually
 * assume that overflows raise errors. `SafeCast` restores this intuition by
 * reverting the transaction when such an operation overflows.
 *
 * Using this library instead of the unchecked operations eliminates an entire
 * class of bugs, so it's recommended to use it always.
 *
 * Can be combined with {SafeMath} and {SignedSafeMath} to extend it to smaller types, by performing
 * all math on `uint256` and `int256` and then downcasting.
'''
class SafeCast:
    '''
     * @dev Returns the downcasted uint224 from uint256, reverting on
     * overflow (when the input is greater than largest uint224).
     *
     * Counterpart to Solidity's `uint224` operator.
     *
     * Requirements:
     *
     * - input must fit into 224 bits
    '''
    def toUint224(value) -> (uint):
        require(value <= uint224.max, "SafeCast: value doesn't fit in 224 bits");
        return uint224(value);

    '''
     * @dev Returns the downcasted uint128 from uint256, reverting on
     * overflow (when the input is greater than largest uint128).
     *
     * Counterpart to Solidity's `uint128` operator.
     *
     * Requirements:
     *
     * - input must fit into 128 bits
    '''
    def toUint128(value) -> (uint):
        require(value <= uint128.max, "SafeCast: value doesn't fit in 128 bits");
        return uint128(value);

    '''
     * @dev Returns the downcasted uint96 from uint256, reverting on
     * overflow (when the input is greater than largest uint96).
     *
     * Counterpart to Solidity's `uint96` operator.
     *
     * Requirements:
     *
     * - input must fit into 96 bits
    '''
    def toUint96(value) -> (uint):
        require(value <= uint96.max, "SafeCast: value doesn't fit in 96 bits");
        return uint96(value);

    '''
     * @dev Returns the downcasted uint64 from uint256, reverting on
     * overflow (when the input is greater than largest uint64).
     *
     * Counterpart to Solidity's `uint64` operator.
     *
     * Requirements:
     *
     * - input must fit into 64 bits
    '''
    def toUint64(value) -> (uint):
        require(value <= uint64.max, "SafeCast: value doesn't fit in 64 bits");
        return uint64(value);

    '''
     * @dev Returns the downcasted uint32 from uint256, reverting on
     * overflow (when the input is greater than largest uint32).
     *
     * Counterpart to Solidity's `uint32` operator.
     *
     * Requirements:
     *
     * - input must fit into 32 bits
    '''
    def toUint32(value) -> (uint):
        require(value <= uint32.max, "SafeCast: value doesn't fit in 32 bits");
        return uint32(value);

    '''
     * @dev Returns the downcasted uint16 from uint256, reverting on
     * overflow (when the input is greater than largest uint16).
     *
     * Counterpart to Solidity's `uint16` operator.
     *
     * Requirements:
     *
     * - input must fit into 16 bits
    '''
    def toUint16(value) -> (uint):
        require(value <= uint16.max, "SafeCast: value doesn't fit in 16 bits");
        return uint16(value);

    '''
     * @dev Returns the downcasted uint8 from uint256, reverting on
     * overflow (when the input is greater than largest uint8).
     *
     * Counterpart to Solidity's `uint8` operator.
     *
     * Requirements:
     *
     * - input must fit into 8 bits.
    '''
    def toUint8(value) -> (uint):
        require(value <= uint8.max, "SafeCast: value doesn't fit in 8 bits");
        return uint8(value);

    '''
     * @dev Converts a signed int256 into an unsigned uint256.
     *
     * Requirements:
     *
     * - input must be greater than or equal to 0.
    '''
    def toUint256(value) -> (uint):
        require(value >= 0, "SafeCast: value must be positive");
        return uint256(value);



# coding=utf-8
# --------------------------------------------------------------------------------------------------------------------
# Licensed under the MIT LICENSE. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------------------------------
"""Mesa Agent-based implementations of the Bancor protocol."""

import mesa
from bancor_research import SECONDS_PER_DAY

from bancor_research.bancor_simulator.v3.spec.actions import (
    unpack_withdrawal_cooldown,
    vortex_burner,
)
from bancor_research.bancor_simulator.v3.spec.network import BancorDapp
from bancor_research.bancor_simulator.v3.simulation.random_walk import RandomWalker
from bancor_research.bancor_simulator.v3.simulation.utils import (
    trade_tkn_to_ema,
    trade_bnt_to_ema,
    process_arbitrage_trade,
)
from bancor_research.bancor_simulator.v3.spec.utils import (
    compute_user_total_holdings,
    compute_ema,
    compute_bntkn_rate,
    compute_max_tkn_deposit,
    compute_master_vault_tkn_tvl,
)


class Trader(RandomWalker):
    """
    Represents a Bancor dapp user (trader and/or arbitrageur). Subclass to Mesa Agent
    """

    def __init__(
        self,
        unique_id: int,
        pos: Tuple[int, int],
        model: mesa.Model,
        moore: bool,
        whale_threshold: Decimal,
        whitelisted_tokens: list,
        target_tvl: Decimal,
    ):

        # init parent class with required parameters
        super().__init__(unique_id, pos, model, moore=moore)

        # init system logging parameters for DataCollector
        self.latest_tkn_name = None
        self.latest_tkn_amt = None
        self.latest_user_name = None
        self.latest_action = None
        self.erc20contracts_bntkn = None
        self.erc20contracts_bnbnt = None
        self.staked_bnt = None
        self.staked_tkn = None
        self.master_vault_bnt = None
        self.master_vault_tkn = None
        self.external_protection_vault_tkn = None
        self.protocol_wallet_bnbnt = None
        self.protocol_wallet_bntkn = None
        self.vortex_bnt = None
        self.ema_rate = None
        self.spot_rate = None

        # all agents use a single BancorDapp instance
        self.protocol = model.protocol

        # init agent attributes
        self.target_tvl = target_tvl
        self.whale_threshold = whale_threshold
        self.user_name = f"user_{unique_id}"
        self.protocol.v3.create_user(self.user_name)

        # the whitelisted tokens available in the model
        for tkn_name in whitelisted_tokens:

            # start everyone off with a random amount in their wallet
            random_amt = Decimal(self.random.randint(1, whale_threshold + 1))
            self.protocol.v3.global_state.set_user_balance(
                user_name=self.user_name, tkn_name=tkn_name, value=random_amt
            )

            if tkn_name == "bnt":
                self.initial_bnt = random_amt
            elif tkn_name == "tkn":
                self.initial_tkn = random_amt

        state = self.protocol.v3.global_state
        self.initial_balance = compute_user_total_holdings(state, self.user_name)
        self.daily_trade_volume = 0

    @property
    def profit(self):
        return (
            compute_user_total_holdings(self.protocol.v3.global_state, self.user_name)
            / self.initial_balance
        )

    def get_random_amt(self, amt: Decimal) -> Decimal:
        if self.random.randint(0, 1000) != 0:
            max_amt, min_amt = amt * Decimal("0.0001"), amt * Decimal("0.000001")
        else:
            max_amt, min_amt = amt * Decimal("0.01"), amt * Decimal("0.001")
        return Decimal(self.random.uniform(float(min_amt), float(max_amt)))

    def perform_random_trade(self):
        """
        Performs a random trade on the server.
        """
        state = self.protocol.v3.global_state
        timestamp = state.timestamp
        user_name = self.user_name
        tkn_name, target_tkn = self.get_random_tkn_names(state)
        source_liquidity = get_tkn_trading_liquidity(state, tkn_name)
        user_funds = get_user_balance(state, user_name, tkn_name)
        swap_amt = self.get_random_amt(source_liquidity)
        if user_funds > swap_amt:
            amt = swap_amt
        else:
            amt = user_funds
        self.protocol.v3.trade(
            tkn_amt=amt,
            source_token=tkn_name,
            target_token=target_tkn,
            user_name=user_name,
            timestamp=timestamp,
        )
        self.daily_trade_volume += amt
        self.latest_tkn_name = tkn_name + "_" + target_tkn
        self.latest_amt = amt
        return self

    def arbitrage_trade(self, state: State, tkn_name: str, user_name: str):
        """
        Computes the appropriate arbitrage trade on the token_name pool.
        """
        timestamp = state.timestamp
        tkn_price, bnt_price = get_prices(state, tkn_name)
        bnt_trading_liquidity = get_bnt_trading_liquidity(state, tkn_name)
        tkn_trading_liquidity = get_tkn_trading_liquidity(state, tkn_name)
        trading_fee = get_trading_fee(state, tkn_name)
        user_tkn = get_user_balance(state, user_name, tkn_name)
        user_bnt = get_user_balance(state, user_name, "bnt")
        is_trading_enabled = get_is_trading_enabled(state, tkn_name)

        if is_trading_enabled:
            x = process_arbitrage_trade(
                tkn_name,
                tkn_price,
                bnt_price,
                bnt_trading_liquidity,
                tkn_trading_liquidity,
                trading_fee,
                user_tkn,
                user_bnt,
            )
            if x is not None:
                (
                    trade_amt,
                    source_token,
                    target_token,
                    user_capability,
                ) = x
                if user_capability:
                    self.protocol.v3.trade(
                        trade_amt, source_token, target_token, user_name, timestamp
                    )
                    self.daily_trade_volume += trade_amt
                self.latest_tkn_name = source_token + "_" + target_token
                self.latest_amt = trade_amt

    def perform_random_arbitrage_trade(self):
        """
        Performs a random arbitrage trade.
        """
        state = self.protocol.v3.global_state
        user_name = self.random.choice(state.usernames)
        tokens = [token for token in state.whitelisted_tokens if token != "bnt"]
        tkn_name = self.random.choice(tokens)
        if get_is_trading_enabled(state, tkn_name):
            self.arbitrage_trade(state, tkn_name, user_name)

        return self

    def process_force_moving_average(
        self, tkn_name: str, user_tkn: Decimal, user_bnt: Decimal
    ) -> Tuple[Decimal, str, str, bool]:
        """
        Process the trade amount to force the ema and the spot price together.
        """
        state = self.protocol.v3.global_state
        tkn_trading_liquidity = get_tkn_trading_liquidity(state, tkn_name)
        bnt_trading_liquidity = get_bnt_trading_liquidity(state, tkn_name)
        trading_fee = get_trading_fee(state, tkn_name)
        network_fee = get_network_fee(state, tkn_name)
        ema_rate = get_ema_rate(state, tkn_name)
        spot_rate = get_spot_rate(state, tkn_name)
        future_ema = compute_ema(spot_rate, ema_rate)
        if ema_rate < spot_rate:
            source_tkn = tkn_name
            target_tkn = "bnt"
            trade_amt = trade_tkn_to_ema(
                bnt_trading_liquidity,
                tkn_trading_liquidity,
                trading_fee,
                network_fee,
                future_ema,
            )
            user_capability = user_tkn > trade_amt
        elif ema_rate > spot_rate:
            source_tkn = "bnt"
            target_tkn = tkn_name
            trade_amt = trade_bnt_to_ema(
                bnt_trading_liquidity,
                tkn_trading_liquidity,
                trading_fee,
                network_fee,
                future_ema,
            )
            user_capability = user_bnt > trade_amt
        else:
            source_tkn = tkn_name
            target_tkn = tkn_name
            trade_amt = Decimal("0")
            user_capability = False

        self.latest_tkn_name = source_tkn + "_" + target_tkn
        self.latest_amt = trade_amt
        return trade_amt, source_tkn, target_tkn, user_capability

    def force_moving_average(
        self, state: State, tkn_name: str, user_name: str, timestamp: int
    ):
        """
        Analyze the state of a pool and creates a swap that drives the ema and the spot price together.
        """
        user_tkn = get_user_balance(state, user_name, tkn_name)
        user_bnt = get_user_balance(state, user_name, "bnt")
        if get_is_trading_enabled(state, tkn_name):
            (
                tkn_amt,
                source_token,
                target_token,
                user_capability,
            ) = self.process_force_moving_average(tkn_name, user_tkn, user_bnt)
            if user_capability:
                self.protocol.v3.trade(
                    tkn_amt, source_token, target_token, user_name, timestamp
                )
                self.daily_trade_volume += tkn_amt
                self.latest_tkn_name = source_token + "_" + target_token
                self.latest_amt = tkn_amt
        return self

    def perform_random_ema_force_trade(self):
        state = self.protocol.v3.global_state
        timestamp = state.timestamp
        user_name = self.random.choice(state.usernames)
        tokens = [token for token in state.whitelisted_tokens if token != "bnt"]
        tkn_name = self.random.choice(tokens)
        if get_is_trading_enabled(state, tkn_name):
            self.force_moving_average(state, tkn_name, user_name, timestamp)

        return self

    def get_random_tkn_names(self, state: State) -> Tuple[str, str]:
        tokens = state.whitelisted_tokens
        source_tkn = self.random.choice(tokens)
        target_tkn = self.random.choice([tkn for tkn in tokens if tkn != source_tkn])
        return source_tkn, target_tkn

    def get_average_trading_fee(self):
        import numpy as np

        state = self.protocol.v3.global_state
        return np.average(
            [
                float(state.tokens[tkn_name].trading_fee)
                for tkn_name in state.whitelisted_tokens
            ]
        )

    def transact(self):
        """
        These probabilities are completely arbitrary.
        """
        trade_motive = self.get_average_trading_fee()
        latest_action = None
        i = self.random.randint(0, 1500)
        band1 = 500 * trade_motive
        if i < band1:
            latest_action = "trade"
            # # print(f'{self.protocol.v3.global_state.timestamp} {self.user_name} perform_random_trade')
            self.perform_random_trade()
        elif band1 <= i < band1 * 2:
            latest_action = "arbitrage_trade"
            # # print(f'{self.protocol.v3.global_state.timestamp} {self.user_name} perform_random_arbitrage_trade')
            self.perform_random_arbitrage_trade()
        elif band1 * 2 <= i < band1 * 3:
            latest_action = "ema_force_trade"
            # print(f'{self.protocol.v3.global_state.timestamp} {self.user_name} perform_random_ema_force_trade')
            self.perform_random_ema_force_trade()

        state = self.protocol.v3.global_state
        self.latest_user_name = self.user_name
        self.latest_action = latest_action
        self.erc20contracts_bntkn = get_pooltoken_balance(state, "tkn")
        self.erc20contracts_bnbnt = get_pooltoken_balance(state, "bnt")
        self.staked_bnt = get_staked_balance(state, "bnt")
        self.staked_tkn = get_staked_balance(state, "tkn")
        self.master_vault_bnt = get_master_vault_balance(state, "bnt")
        self.master_vault_tkn = get_master_vault_balance(state, "tkn")
        self.external_protection_vault_tkn = get_external_protection_vault_balance(state, "tkn")
        self.protocol_wallet_bnbnt = get_protocol_wallet_balance(state, "bnt")
        self.protocol_wallet_bntkn = get_protocol_wallet_balance(state, "tkn")
        self.vortex_bnt = get_vortex_balance(state, "bnt")
        self.ema_rate = get_ema_rate(state, "tkn")
        self.spot_rate = get_spot_rate(state, "tkn")

    # step is called for each agent in model.BancorSimulator.schedule.step()
    def step(self):
        # move to a cell in my Moore neighborhood
        self.random_move()
        self.transact()


class LP(RandomWalker):
    """
    Represents a Bancor dapp liquidity provider. Subclass to Mesa Agent
    """

    def __init__(
        self,
        unique_id: int,
        pos: Tuple[int, int],
        model: mesa.Model,
        moore: bool,
        whale_threshold: Decimal,
        whitelisted_tokens: list,
        target_tvl: Decimal,
    ):

        # init parent class with required parameters
        super().__init__(unique_id, pos, model, moore=moore)

        # Bancor dapp, set at __init__, all people use the same dapp
        self.latest_tkn_name = None
        self.latest_tkn_amt = None
        self.latest_user_name = None
        self.latest_action = None
        self.erc20contracts_bntkn = None
        self.erc20contracts_bnbnt = None
        self.staked_bnt = None
        self.staked_tkn = None
        self.master_vault_bnt = None
        self.master_vault_tkn = None
        self.external_protection_vault_tkn = None
        self.protocol_wallet_bnbnt = None
        self.protocol_wallet_bntkn = None
        self.vortex_bnt = None
        self.ema_rate = None
        self.spot_rate = None

        user_name = f"user_{unique_id}"
        self.whale_threshold = whale_threshold

        self.protocol = model.protocol
        self.protocol.v3.create_user(user_name)
        self.user_name = user_name
        self.target_tvl = Decimal("100000000")

        # the whitelisted tokens available in the model
        for tkn_name in self.protocol.v3.global_state.whitelisted_tokens:

            # start everyone off with a random amount in their wallet
            random_amt = Decimal(self.random.randint(1, whale_threshold + 1))
            self.protocol.v3.global_state.set_user_balance(
                user_name=self.user_name, tkn_name=tkn_name, value=random_amt
            )

            if tkn_name == "bnt":
                self.initial_bnt = random_amt
            elif tkn_name == "tkn":
                self.initial_tkn = random_amt

        state = self.protocol.v3.global_state
        self.initial_balance = compute_user_total_holdings(state, self.user_name)

    @property
    def profit(self):
        return (
            compute_user_total_holdings(self.protocol.v3.global_state, self.user_name)
            / self.initial_balance
        )

    def get_random_trading_fee(self, trading_fee: Decimal) -> Decimal:
        min_trading_fee, max_trading_fee = max(
            trading_fee / Decimal("3"), Decimal("0.001")
        ), min(trading_fee * Decimal("3"), Decimal("0.05"))
        return Decimal(
            self.random.uniform(float(min_trading_fee), float(max_trading_fee))
        )

    def get_random_network_fee(self, network_fee: Decimal) -> Decimal:
        min_fee, max_fee = max(network_fee / Decimal("2"), Decimal("0.05")), min(
            network_fee * Decimal("2"), Decimal("0.25")
        )
        return Decimal(self.random.uniform(float(min_fee), float(max_fee)))

    def get_random_withdrawal_fee(self, withdrawal_fee: Decimal) -> Decimal:
        min_fee, max_fee = max(withdrawal_fee / Decimal("2"), Decimal("0.001")), min(
            withdrawal_fee * Decimal("2"), Decimal("0.01")
        )
        return Decimal(self.random.uniform(float(min_fee), float(max_fee)))

    def get_random_bnt_funding_limit(self, bnt_funding_amt: Decimal) -> Decimal:
        min_bnt_funding_limit = bnt_funding_amt * Decimal("1.5")
        max_bnt_funding_limit = bnt_funding_amt * Decimal("3.0")
        bnt_funding_limit = Decimal(
            self.random.uniform(
                float(min_bnt_funding_limit), float(max_bnt_funding_limit)
            )
        )
        return bnt_funding_limit

    def set_random_trading_fee(self):
        state = self.protocol.v3.global_state
        for i in range(self.random.randint(1, 3)):
            tkn_name = self.random.choice(state.whitelisted_tokens)
            trading_fee = get_trading_fee(state, tkn_name)
            trading_fee = self.get_random_trading_fee(trading_fee)
            state.set_trading_fee(tkn_name, trading_fee)
            self.protocol.v3.set_state(state)
        return self

    def set_random_network_fee(self):
        state = self.protocol.v3.global_state
        tkn_name = self.random.choice(state.whitelisted_tokens)
        network_fee = get_network_fee(state, tkn_name)
        network_fee = self.get_random_network_fee(network_fee)
        state.set_network_fee(tkn_name, network_fee)
        self.protocol.v3.set_state(state)
        return self

    def set_random_withdrawal_fee(self):
        state = self.protocol.v3.global_state
        withdrawal_fee = state.withdrawal_fee
        tkn_name = self.random.choice(state.whitelisted_tokens)
        withdrawal_fee = self.get_random_withdrawal_fee(withdrawal_fee)
        state.set_withdrawal_fee(tkn_name, withdrawal_fee)
        self.protocol.v3.set_state(state)
        return self

    def set_random_bnt_funding_limit(self):
        state = self.protocol.v3.global_state
        tkn_name = self.random.choice(state.whitelisted_tokens)
        bnt_funding_limit = get_bnt_funding_limit(state, tkn_name)
        updated_bnt_funding_limit = self.get_random_bnt_funding_limit(bnt_funding_limit)
        state.set_bnt_funding_limit(tkn_name, updated_bnt_funding_limit)
        self.protocol.v3.set_state(state)
        return self

    def perform_random_enable_trading(self):
        for tkn_name in self.protocol.v3.global_state.whitelisted_tokens:
            self.protocol.v3.enable_trading(tkn_name)
        return self

    def get_random_withdrawal_amt(self, tkn_name: str) -> Decimal:
        user_balance = get_user_balance(
            self.protocol.v3.global_state, self.user_name, tkn_name
        )
        return user_balance * Decimal(self.random.uniform(float(0.001), float(0.5)))

    def get_random_cooldown_amt(self, user_bntkn_amt: Decimal) -> Decimal:
        if self.random.randint(0, 10) != 0:
            max_amt, min_amt = user_bntkn_amt * Decimal(
                "0.1"
            ), user_bntkn_amt * Decimal("0.01")
        else:
            max_amt, min_amt = user_bntkn_amt * Decimal("1"), user_bntkn_amt * Decimal(
                "0.5"
            )
        return Decimal(self.random.uniform(float(min_amt), float(max_amt)))

    def get_random_amt(self, amt: Decimal) -> Decimal:
        if self.random.randint(0, 1000) != 0:
            max_amt, min_amt = amt * Decimal("0.0001"), amt * Decimal("0.000001")
        else:
            max_amt, min_amt = amt * Decimal("0.01"), amt * Decimal("0.001")
        return Decimal(self.random.uniform(float(min_amt), float(max_amt)))

    def get_random_tkn_names(self, state: State) -> Tuple[str, str]:
        tokens = state.whitelisted_tokens
        source_tkn = self.random.choice(tokens)
        target_tkn = self.random.choice([tkn for tkn in tokens if tkn != source_tkn])
        return source_tkn, target_tkn

    def is_protocol_bnbnt_healthy(
        self, protocol_bnbnt: Decimal, bnbnt_supply: Decimal
    ) -> bool:
        """
        Checks if the protocol owned bnbnt is at a healthy level (greater than 50%)
        """
        return protocol_bnbnt / bnbnt_supply > Decimal("0.5")

    def get_deposit_payload(
        self, state: State
    ) -> Tuple[str, str, Decimal, Decimal, Decimal, Decimal, Decimal]:
        """
        Gets the input data required for a deposit.
        """
        user_name = self.user_name
        tkn_name, target_tkn = self.get_random_tkn_names(state)
        user_tkn = get_user_balance(state, user_name, tkn_name)
        user_bnt = get_user_balance(state, user_name, "bnt")
        bnbnt_supply = get_pooltoken_balance(state, "bnt")
        protocol_bnbnt = get_protocol_wallet_balance(state, "bnt")
        bnbnt_rate = get_bnbnt_rate(state)
        return (
            user_name,
            tkn_name,
            user_tkn,
            user_bnt,
            bnbnt_supply,
            protocol_bnbnt,
            bnbnt_rate,
        )

    def process_random_autocompounding_program(self):
        """
        Performs a random trade on the server.
        """
        state = self.protocol.v3.global_state
        timestamp = state.timestamp * 100000

        for tkn_name in state.whitelisted_tokens:
            if tkn_name in state.active_autocompounding_programs:
                self.protocol.v3.process_ac_rewards_program(
                    tkn_name=tkn_name, timestamp=timestamp
                )

    def create_random_autocompounding_rewards(self):
        """
        Performs a random trade on the server.
        """
        state = self.protocol.v3.global_state
        timestamp = state.timestamp
        start_time = 1 + timestamp

        tkn_name = self.random.choice(state.whitelisted_tokens)
        distribution_type = self.random.choice(["flat", "exp"])

        if distribution_type == "flat":
            self.protocol.v3.create_flat_ac_rewards_program(
                tkn_name=tkn_name,
                user_name=self.user_name,
                total_rewards="86400",
                start_time=start_time,
                total_duration=365*SECONDS_PER_DAY,
                timestamp=timestamp,
            )
        else:
            self.protocol.v3.create_exp_ac_rewards_program(
                tkn_name=tkn_name,
                user_name=self.user_name,
                total_rewards="360000",
                start_time=start_time,
                half_life=1*SECONDS_PER_DAY,
                timestamp=timestamp,
            )

    def perform_random_trade(self):
        """
        Performs a random trade on the server.
        """
        state = self.protocol.v3.global_state
        timestamp = state.timestamp
        for i in range(self.random.randint(1, 10)):
            user_name = self.user_name
            tkn_name, target_tkn = self.get_random_tkn_names(state)
            source_liquidity = get_tkn_trading_liquidity(state, tkn_name)
            user_funds = get_user_balance(state, user_name, tkn_name)
            for i in range(self.random.randint(1, 5)):
                swap_amt = self.get_random_amt(source_liquidity)
                if user_funds > swap_amt:
                    amt = swap_amt
                else:
                    amt = user_funds
                self.protocol.v3.trade(
                    tkn_amt=amt,
                    source_token=tkn_name,
                    target_token=target_tkn,
                    user_name=user_name,
                    timestamp=timestamp,
                )

        return self

    def arbitrage_trade(self, state: State, tkn_name: str, user_name: str):
        """
        Computes the appropriate arbitrage trade on the token_name pool.
        """
        timestamp = state.timestamp
        tkn_price, bnt_price = get_prices(state, tkn_name)
        bnt_trading_liquidity = get_bnt_trading_liquidity(state, tkn_name)
        tkn_trading_liquidity = get_tkn_trading_liquidity(state, tkn_name)
        trading_fee = get_trading_fee(state, tkn_name)
        user_tkn = get_user_balance(state, user_name, tkn_name)
        user_bnt = get_user_balance(state, user_name, "bnt")
        is_trading_enabled = get_is_trading_enabled(state, tkn_name)

        if is_trading_enabled:
            x = process_arbitrage_trade(
                tkn_name,
                tkn_price,
                bnt_price,
                bnt_trading_liquidity,
                tkn_trading_liquidity,
                trading_fee,
                user_tkn,
                user_bnt,
            )
            if x is not None:
                (
                    trade_amt,
                    source_token,
                    target_token,
                    user_capability,
                ) = x
                if user_capability:
                    self.protocol.v3.trade(
                        trade_amt, source_token, target_token, user_name, timestamp
                    )

    def perform_random_begin_cooldown(self):
        """
        Begins a random cooldown.
        """
        state = self.protocol.v3.global_state
        timestamp = state.timestamp
        for i in range(self.random.randint(1, 10)):
            user_name = self.user_name
            tkn_name, target_tkn = self.get_random_tkn_names(state)
            user_bntkn_amt = get_user_balance(state, user_name, tkn_name)
            bntkn_rate = compute_bntkn_rate(state, tkn_name)
            if user_bntkn_amt > 0 and bntkn_rate > 0:
                bntkn_amt = self.get_random_cooldown_amt(user_bntkn_amt)
                withdraw_value = bntkn_amt / bntkn_rate
                self.protocol.v3.begin_cooldown_by_rtkn(
                    tkn_name=tkn_name,
                    tkn_amt=withdraw_value,
                    user_name=user_name,
                    timestamp=timestamp,
                )

        return self

    def perform_random_arbitrage_trade(self):
        """
        Performs a random arbitrage trade.
        """
        state = self.protocol.v3.global_state
        for i in range(self.random.randint(1, 10)):
            user_name = self.random.choice(state.usernames)
            tokens = [token for token in state.whitelisted_tokens if token != "bnt"]
            tkn_name = self.random.choice(tokens)
            if get_is_trading_enabled(state, tkn_name):
                self.arbitrage_trade(state, tkn_name, user_name)

        return self

    def perform_random_deposit(self):
        """
        Performs a random deposit action
        """
        target_tvl = self.target_tvl
        state = self.protocol.v3.global_state
        timestamp = state.timestamp
        (
            user_name,
            tkn_name,
            user_tkn,
            user_bnt,
            bnbnt_supply,
            protocol_bnbnt,
            bnbnt_rate,
        ) = self.get_deposit_payload(state)
        deposit_amt = None
        if tkn_name != "bnt":
            master_vault_balance = get_master_vault_balance(state, tkn_name)
            token_price, bnt_price = get_prices(state, tkn_name)
            master_vault_tvl = compute_master_vault_tkn_tvl(master_vault_balance, token_price)
            if master_vault_tvl < target_tvl:
                max_tkn_deposit = compute_max_tkn_deposit(
                    master_vault_tvl, target_tvl, user_tkn
                )
                deposit_amt = self.get_random_amt(max_tkn_deposit)
                if deposit_amt < user_tkn:
                    self.protocol.v3.deposit(
                        tkn_name,
                        deposit_amt,
                        user_name,
                        timestamp,
                    )
        elif tkn_name == "bnt":
            if bnbnt_supply > 0 and self.is_protocol_bnbnt_healthy(
                protocol_bnbnt, bnbnt_supply
            ):
                max_bnt_deposit = get_max_bnt_deposit(
                    bnbnt_supply, protocol_bnbnt, bnbnt_rate, user_bnt
                )
                deposit_amt = self.get_random_amt(max_bnt_deposit)
                if deposit_amt < user_tkn:
                    self.protocol.v3.deposit(
                        tkn_name, deposit_amt, user_name, timestamp
                    )

        self.latest_tkn_name = tkn_name
        self.latest_amt = deposit_amt
        return self

    def process_withdrawal(
        self,
        user_name: str,
        id_number: int,
        timestamp: int = 0,
    ):
        """
        Main withdrawal logic based on the withdraw algorithm of the BIP15 spec.
        """
        state = self.protocol.v3.global_state

        (
            id_number,
            cooldown_timestamp,
            tkn_name,
            pool_token_amt,
            withdraw_value,
            user_name,
        ) = unpack_withdrawal_cooldown(state, user_name, id_number)

        cooldown_time = state.cooldown_time
        cool_down_complete = timestamp - cooldown_timestamp >= cooldown_time
        if cool_down_complete:
            self.protocol.v3.withdraw(
                user_name=user_name,
                id_number=id_number,
                timestamp=timestamp,
                tkn_name=tkn_name,
                tkn_amt=withdraw_value,
            )
        self.latest_tkn_name = tkn_name
        self.latest_amt = withdraw_value
        return self

    def perform_random_withdrawal(self):
        """
        Perform random withdraw action.
        """
        state = self.protocol.v3.global_state
        timestamp = state.timestamp
        user_name = self.random.choice(state.usernames)
        tkn_name = self.random.choice(state.whitelisted_tokens)
        pending_withdrawals = get_user_pending_withdrawals(state, user_name, tkn_name)
        if len(pending_withdrawals) > 0:
            id_number = self.random.choice(pending_withdrawals)
            self.process_withdrawal(
                user_name=user_name, id_number=id_number, timestamp=timestamp
            )
        return self

    def process_force_moving_average(
        self, tkn_name: str, user_tkn: Decimal, user_bnt: Decimal
    ) -> Tuple[Decimal, str, str, bool]:
        """
        Process the trade amount to force the ema and the spot price together.
        """
        state = self.protocol.v3.global_state
        tkn_trading_liquidity = get_tkn_trading_liquidity(state, tkn_name)
        bnt_trading_liquidity = get_bnt_trading_liquidity(state, tkn_name)
        trading_fee = get_trading_fee(state, tkn_name)
        network_fee = get_network_fee(state, tkn_name)
        ema_rate = get_ema_rate(state, tkn_name)
        spot_rate = get_spot_rate(state, tkn_name)
        future_ema = compute_ema(spot_rate, ema_rate)
        if ema_rate < spot_rate:
            source_tkn = tkn_name
            target_tkn = "bnt"
            trade_amt = trade_tkn_to_ema(
                bnt_trading_liquidity,
                tkn_trading_liquidity,
                trading_fee,
                network_fee,
                future_ema,
            )
            user_capability = user_tkn > trade_amt
        elif ema_rate > spot_rate:
            source_tkn = "bnt"
            target_tkn = tkn_name
            trade_amt = trade_bnt_to_ema(
                bnt_trading_liquidity,
                tkn_trading_liquidity,
                trading_fee,
                network_fee,
                future_ema,
            )
            user_capability = user_bnt > trade_amt
        else:
            source_tkn = tkn_name
            target_tkn = tkn_name
            trade_amt = Decimal("0")
            user_capability = False
        self.latest_tkn_name = source_tkn + "_" + target_tkn
        self.latest_amt = trade_amt
        return trade_amt, source_tkn, target_tkn, user_capability

    def force_moving_average(
        self, state: State, tkn_name: str, user_name: str, timestamp: int
    ):
        """
        Analyze the state of a pool and creates a swap that drives the ema and the spot price together.
        """
        user_tkn = get_user_balance(state, user_name, tkn_name)
        user_bnt = get_user_balance(state, user_name, "bnt")
        if get_is_trading_enabled(state, tkn_name):
            (
                tkn_amt,
                source_token,
                target_token,
                user_capability,
            ) = self.process_force_moving_average(tkn_name, user_tkn, user_bnt)
            if user_capability:
                self.protocol.v3.trade(
                    tkn_amt, source_token, target_token, user_name, timestamp
                )
                self.latest_tkn_name = source_token + "_" + target_token
                self.latest_amt = tkn_amt
        return self

    def perform_random_ema_force_trade(self):
        state = self.protocol.v3.global_state
        timestamp = state.timestamp
        user_name = self.random.choice(state.usernames)
        tokens = [token for token in state.whitelisted_tokens if token != "bnt"]
        tkn_name = self.random.choice(tokens)
        if get_is_trading_enabled(state, tkn_name):
            self.force_moving_average(state, tkn_name, user_name, timestamp)

        return self

    def perform_random_vortex_burner(self):
        """
        Simulation purposes only.
        """
        state = self.protocol.v3.global_state
        vortex_burner(state, self.user_name)
        return self

    def transact(self):
        """
        These probabilities are completely arbitrary.
        """

        i = self.random.randint(0, 1242)
        if i < 400:
            latest_action = "deposit"
            # print(f'{self.protocol.v3.global_state.timestamp} {self.user_name} perform_random_deposit')
            self.perform_random_deposit()
        elif 400 <= i < 500:
            latest_action = "create_autocompounding"
            # print(f'{self.protocol.v3.global_state.timestamp} {self.user_name} create_random_autocompounding_rewards')
            self.create_random_autocompounding_rewards()
        elif 500 <= i < 600:
            latest_action = "process_autocompounding"
            # print(f'{self.protocol.v3.global_state.timestamp} {self.user_name} process_random_autocompounding_program')
            self.process_random_autocompounding_program()
        elif 600 <= i < 1000:
            latest_action = "deposit"
            # print(f'{self.protocol.v3.global_state.timestamp} {self.user_name} perform_random_deposit')
            self.perform_random_deposit()
        elif 1000 <= i < 1100:
            latest_action = "cooldown"
            # print(f'{self.protocol.v3.global_state.timestamp} {self.user_name} perform_random_begin_cooldown')
            self.perform_random_begin_cooldown()
        elif 1100 <= i < 1200:
            latest_action = "withdrawal"
            # print(f'{self.protocol.v3.global_state.timestamp} {self.user_name} perform_random_withdrawal')
            self.perform_random_withdrawal()
        elif 1200 <= i < 1210:
            latest_action = "vortex_burner"
            # print(f'{self.protocol.v3.global_state.timestamp} {self.user_name} perform_random_vortex_burner')
            self.perform_random_vortex_burner()
        elif 1210 <= i < 1215:
            latest_action = "change_trading_fee"
            # print(f'{self.protocol.v3.global_state.timestamp} {self.user_name} set_random_trading_fee')
            self.set_random_trading_fee()
        elif 1215 <= i < 1220:
            latest_action = "change_network_fee"
            # print(f'{self.protocol.v3.global_state.timestamp} {self.user_name} set_random_network_fee')
            self.set_random_network_fee()
        elif 1220 <= i < 1230:
            latest_action = "change_withdrawal_fee"
            # print(f'{self.protocol.v3.global_state.timestamp} {self.user_name} set_random_withdrawal_fee')
            self.set_random_withdrawal_fee()
        elif 1230 <= i < 1240:
            latest_action = "change_bnt_funding_limit"
            # print(f'{self.protocol.v3.global_state.timestamp} {self.user_name} set_random_bnt_funding_limit')
            self.set_random_bnt_funding_limit()
        elif 1240 <= i <= 1242:
            latest_action = "enable_trading"
            # print(f'{self.protocol.v3.global_state.timestamp} {self.user_name} perform_random_enable_trading')
            self.perform_random_enable_trading()

        state = self.protocol.v3.global_state
        self.latest_user_name = self.user_name
        self.latest_action = latest_action
        self.erc20contracts_bntkn = get_pooltoken_balance(state, "tkn")
        self.erc20contracts_bnbnt = get_pooltoken_balance(state, "bnt")
        self.staked_bnt = get_staked_balance(state, "bnt")
        self.staked_tkn = get_staked_balance(state, "tkn")
        self.master_vault_bnt = get_master_vault_balance(state, "bnt")
        self.master_vault_tkn = get_master_vault_balance(state, "tkn")
        self.external_protection_vault_tkn = get_external_protection_vault_balance(state, "tkn")
        self.protocol_wallet_bnbnt = get_protocol_wallet_balance(state, "bnt")
        self.protocol_wallet_bntkn = get_protocol_wallet_balance(state, "tkn")
        self.vortex_bnt = get_vortex_balance(state, "bnt")
        self.ema_rate = get_ema_rate(state, "tkn")
        self.spot_rate = get_spot_rate(state, "tkn")

    # step is called for each agent in model.BancorSimulator.schedule.step()
    def step(self):
        # move to a cell in my Moore neighborhood
        self.random_move()
        self.transact()


class Protocol(mesa.Agent):
    def __init__(
        self,
        unique_id,
        model,
        trading_fee,
        network_fee,
        withdrawal_fee,
        whitelisted_tokens,
        price_feeds,
        bnt_funding_limit,
        bnt_min_liquidity,
    ):

        # initialize the parent class with required parameters
        super().__init__(unique_id, model)

        # set the dapp interface to Bancor v3
        self.v3 = BancorDapp(
            price_feeds=price_feeds,
            whitelisted_tokens=whitelisted_tokens,
            trading_fee=trading_fee,
            network_fee=network_fee,
            withdrawal_fee=withdrawal_fee,
            bnt_funding_limit=bnt_funding_limit,
            bnt_min_liquidity=bnt_min_liquidity,
        )

    @property
    def bntkn_health(self) -> Decimal:
        state = self.v3.global_state
        pooltoken_balance = get_pooltoken_balance(state, "tkn")
        protocol_bntkn = get_protocol_wallet_balance(state, "tkn")
        if pooltoken_balance > Decimal("0"):
            return protocol_bntkn / pooltoken_balance
        else:
            return Decimal("0")

    @property
    def bnbnt_health(self) -> Decimal:
        state = self.v3.global_state
        pooltoken_balance = get_pooltoken_balance(state, "bnt")
        protocol_bnbnt = get_protocol_wallet_balance(state, "bnt")
        if pooltoken_balance > Decimal("0"):
            return protocol_bnbnt / pooltoken_balance
        else:
            return Decimal("0")



# coding=utf-8
# --------------------------------------------------------------------------------------------------------------------
# Licensed under the MIT LICENSE. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------------------------------
"""Sets up the interactive visualization server for the Mesa Agent-based simulation."""

# The colors here are taken from Matplotlib's tab10 palette
RICH_COLOR = "#2ca02c"  # Green
POOR_COLOR = "#d62728"  # Red
MID_COLOR = "#1f77b4"  # Blue
INIT_TRADER_COLOR = "yellow"
INIT_LP_COLOR = "skyblue"


def agent_portrayal(agent: mesa.Agent):
    """
    A CanvasGrid object uses a user-provided portrayal method to generate a
    portrayal for each object. A portrayal is a JSON-ready dictionary which
    tells the relevant JavaScript code (GridDraw.js) where to draw what shape.

    The render method returns a dictionary, keyed on layers, with values as
    lists of portrayals to draw. Portrayals themselves are generated by the
    user-provided portrayal_method, which accepts an object as an input and
    produces a portrayal of it.

    A portrayal as a dictionary with the following structure:
        "x", "y": Coordinates for the cell in which the object is placed.
        "Shape": Can be either "circle", "rect", "arrowHead" or a custom image.
            For Circles:
                "r": The radius, defined as a fraction of cell size. r=1 will
                     fill the entire cell.
                "xAlign", "yAlign": Alignment of the circle within the cell.
                                    Defaults to 0.5 (center).
            For Rectangles:
                "w", "h": The width and height of the rectangle, which are in
                          fractions of cell width and height.
                "xAlign", "yAlign": Alignment of the rectangle within the
                                    cell. Defaults to 0.5 (center).
            For arrowHead:
                "scale": Proportion scaling as a fraction of cell size.
                "heading_x": represents x direction unit vector.
                "heading_y": represents y direction unit vector.
             For an image:
                The image must be placed in the same directory from which the
                server is launched. An image has the attributes "x", "y",
                "scale", "text" and "text_color".
        "Color": The color to draw the shape in; needs to be a valid HTML
                 color, e.g."Red" or "#AA08F8"
        "Filled": either "true" or "false", and determines whether the shape is
                  filled or not.
        "Layer": Layer number of 0 or above; higher-numbered layers are drawn
                 above lower-numbered layers.
        "text": The text to be inscribed inside the Shape. Normally useful for
                showing the unique_id of the agent.
        "text_color": The color to draw the inscribed text. Should be given in
                      conjunction of "text" property.


    Attributes:
        portrayal_method: Function which generates portrayals from objects, as
                          described above.
        grid_height, grid_width: Size of the grid to visualize, in cells.
        canvas_height, canvas_width: Size, in pixels, of the grid visualization
                                     to draw on the client.
        template: "canvas_module.html" stores the module's HTML template.

    """
    if agent is None:
        return

    portrayal = {}

    # update portrayal characteristics for each LP object
    if isinstance(agent, LP):
        portrayal["Shape"] = "circle"
        portrayal["r"] = 0.5
        portrayal["Layer"] = 0
        portrayal["Filled"] = "true"

        color = INIT_LP_COLOR

        print("LP ", agent.user_name, agent.profit)

        # set agent color based on savings and loans
        threshold = 90
        lower = 10
        if threshold < agent.profit * 100:
            color = RICH_COLOR
        elif Decimal(lower) < agent.profit * 100 < Decimal(threshold):
            color = MID_COLOR
        elif agent.profit * 100 < Decimal(lower):
            color = POOR_COLOR

        portrayal["Color"] = color

    elif isinstance(agent, Trader):
        portrayal["Shape"] = "rect"
        portrayal["w"] = 0.5
        portrayal["h"] = 0.5
        portrayal["Layer"] = 1
        portrayal["Filled"] = "true"

        color = INIT_TRADER_COLOR

        print("Trader ", agent.user_name, agent.profit)

        # set agent color based on savings and loans
        threshold = 90
        lower = 10
        if threshold < agent.profit * 100:
            color = RICH_COLOR
        elif Decimal(lower) < agent.profit * 100 < Decimal(threshold):
            color = MID_COLOR
        elif agent.profit * 100 < Decimal(lower):
            color = POOR_COLOR

        portrayal["Color"] = color

    return portrayal


# dictionary of user settable parameters - these map to the model __init__ parameters
model_params = {
    "whale_threshold": mesa.visualization.Slider(
        name="Whale Threshold",
        value=float(SIMULATION_WHALE_THRESHOLD),
        min_value=1000,
        max_value=200000,
        step=1000,
        description="Upper End of Random Initial Wallet Amount",
    ),
    "trading_fee": mesa.visualization.Slider(
        name="Trading Fee",
        value=float(DEFAULT_TRADING_FEE),
        min_value=0.001,
        max_value=0.05,
        step=0.001,
        description="Trading Fee",
    ),
    "network_fee": mesa.visualization.Slider(
        name="Network Fee",
        value=float(DEFAULT_NETWORK_FEE),
        min_value=0.001,
        max_value=0.05,
        step=0.001,
        description="Network fee",
    ),
    "withdrawal_fee": mesa.visualization.Slider(
        name="Withdrawal Fee",
        value=float(DEFAULT_WITHDRAWAL_FEE),
        min_value=0.001,
        max_value=0.05,
        step=0.001,
        description="Withdrawal fee",
    ),
}

# set the portrayal function and size of the canvas for visualization
canvas_element = mesa.visualization.CanvasGrid(
    portrayal_method=agent_portrayal,
    grid_width=20,
    grid_height=20,
    # canvas_width=500, canvas_height=500
)

# map data to chart in the ChartModule
chart_element = mesa.visualization.ChartModule(
    [
        {"Label": "Balance > 90%", "Color": RICH_COLOR},
        {"Label": "Balance < 10%", "Color": POOR_COLOR},
        {"Label": "10% < Balance < 90%", "Color": MID_COLOR},
    ]
)

# create instance of Mesa ModularServer
server = mesa.visualization.ModularServer(
    visualization_elements=[canvas_element, chart_element],
    name="Bancor Simulation Model",
    model_params=model_params,
    model_cls=BancorSimulation,
)



from bancor_research.bancor_emulator.solidity import uint, uint8, uint16, uint32, uint128, uint256, mapping, address, payable, revert
from bancor_research.bancor_emulator.utils import contract, using, parse

from bancor_research.bancor_emulator.EnumerableSet import EnumerableSet
from bancor_research.bancor_emulator.Math import Math
from bancor_research.bancor_emulator.SafeCast import SafeCast
from bancor_research.bancor_emulator.Constants import PPM_RESOLUTION
from bancor_research.bancor_emulator.BlockNumber import BlockNumber
from bancor_research.bancor_emulator.FractionLibrary import Fraction256, Fraction112, FractionLibrary, zeroFraction112
from bancor_research.bancor_emulator.MathEx import Sint256, MathEx
from bancor_research.bancor_emulator.PoolToken import PoolToken as IPoolToken
from bancor_research.bancor_emulator.PoolCollectionWithdrawal import PoolCollectionWithdrawal

class PoolLiquidity:
    def __init__(self, x = None) -> None:
        self.bntTradingLiquidity = parse(uint128, x, 'bntTradingLiquidity'); # the BNT trading liquidity
        self.baseTokenTradingLiquidity = parse(uint128, x, 'baseTokenTradingLiquidity'); # the base token trading liquidity
        self.stakedBalance = parse(uint256, x, 'stakedBalance'); # the staked balance

class AverageRates:
    def __init__(self, x = None) -> None:
        self.blockNumber = parse(uint32, x, 'blockNumber');
        self.rate = parse(Fraction112, x, 'rate');
        self.invRate = parse(Fraction112, x, 'invRate');

class Pool:
    def __init__(self, x = None) -> None:
        self.poolToken = parse(address, x, 'poolToken'); # the pool token of the pool
        self.tradingFeePPM = parse(uint32, x, 'tradingFeePPM'); # the trading fee (in units of PPM)
        self.tradingEnabled = parse(bool, x, 'tradingEnabled'); # whether trading is enabled
        self.depositingEnabled = parse(bool, x, 'depositingEnabled'); # whether depositing is enabled
        self.averageRates = parse(AverageRates, x, 'averageRates'); # the recent average rates
        self.liquidity = parse(PoolLiquidity, x, 'liquidity'); # the overall liquidity in the pool

class WithdrawalAmounts:
    def __init__(self, x = None) -> None:
        self.totalAmount = parse(uint256, x, 'totalAmount');
        self.baseTokenAmount = parse(uint256, x, 'baseTokenAmount');
        self.bntAmount = parse(uint256, x, 'bntAmount');

# trading enabling/disabling reasons
TRADING_STATUS_UPDATE_DEFAULT = uint8(0);
TRADING_STATUS_UPDATE_ADMIN = uint8(1);
TRADING_STATUS_UPDATE_MIN_LIQUIDITY = uint8(2);
TRADING_STATUS_UPDATE_INVALID_STATE = uint8(3);

class TradeAmountAndFee:
    def __init__(self, x = None) -> None:
        self.amount = parse(uint256, x, 'amount'); # the source/target amount (depending on the context) resulting from the trade
        self.tradingFeeAmount = parse(uint256, x, 'tradingFeeAmount'); # the trading fee amount
        self.networkFeeAmount = parse(uint256, x, 'networkFeeAmount'); # the network fee amount (always in units of BNT)

# base token withdrawal output amounts
class InternalWithdrawalAmounts:
    def __init__(self, x = None) -> None:
        self.baseTokensToTransferFromMasterVault = parse(uint256, x, 'baseTokensToTransferFromMasterVault'); # base token amount to transfer from the master vault to the provider
        self.bntToMintForProvider = parse(uint256, x, 'bntToMintForProvider'); # BNT amount to mint directly for the provider
        self.baseTokensToTransferFromEPV = parse(uint256, x, 'baseTokensToTransferFromEPV'); # base token amount to transfer from the external protection vault to the provider
        self.baseTokensTradingLiquidityDelta = parse(Sint256, x, 'baseTokensTradingLiquidityDelta'); # base token amount to add to the trading liquidity
        self.bntTradingLiquidityDelta = parse(Sint256, x, 'bntTradingLiquidityDelta'); # BNT amount to add to the trading liquidity and to the master vault
        self.bntProtocolHoldingsDelta = parse(Sint256, x, 'bntProtocolHoldingsDelta'); # BNT amount add to the protocol equity
        self.baseTokensWithdrawalFee = parse(uint256, x, 'baseTokensWithdrawalFee'); # base token amount to keep in the pool as a withdrawal fee
        self.baseTokensWithdrawalAmount = parse(uint256, x, 'baseTokensWithdrawalAmount'); # base token amount equivalent to the base pool token's withdrawal amount
        self.poolTokenAmount = parse(uint256, x, 'poolTokenAmount'); # base pool token
        self.poolTokenTotalSupply = parse(uint256, x, 'poolTokenTotalSupply'); # base pool token's total supply
        self.newBaseTokenTradingLiquidity = parse(uint256, x, 'newBaseTokenTradingLiquidity'); # new base token trading liquidity
        self.newBNTTradingLiquidity = parse(uint256, x, 'newBNTTradingLiquidity'); # new BNT trading liquidity

class TargetTradingLiquidity:
    def __init__(self, x = None) -> None:
        self.update = parse(bool, x, 'update');
        self.bnt = parse(uint256, x, 'bnt');
        self.baseToken = parse(uint256, x, 'baseToken');

class PoolRateState:
    Uninitialized = 0
    Unstable = 1
    Stable = 2

'''
 * @dev Pool Collection contract
 *
 * notes:
 *
 * - the address of reserve token serves as the pool unique ID in both contract functions and events
'''
class PoolCollection(contract, BlockNumber):
    using(FractionLibrary, Fraction256);
    using(FractionLibrary, Fraction112);
    using(SafeCast, uint);

    POOL_TYPE = uint16(1);
    LIQUIDITY_GROWTH_FACTOR = uint256(2);
    BOOTSTRAPPING_LIQUIDITY_BUFFER_FACTOR = uint256(2);
    DEFAULT_TRADING_FEE_PPM = uint32(2_000); # 0.2%
    DEFAULT_NETWORK_FEE_PPM = uint32(200_000); # 20%
    RATE_MAX_DEVIATION_PPM = uint32(10_000); # %1
    RATE_RESET_BLOCK_THRESHOLD = uint32(100);

    # the average rate is recalculated based on the ratio between the weights of the rates the smaller the weights are,
    # the larger the supported range of each one of the rates is
    EMA_AVERAGE_RATE_WEIGHT = uint256(4);
    EMA_SPOT_RATE_WEIGHT = uint256(1);

    class TradeIntermediateResult:
        def __init__(self, x = None) -> None:
            self.sourceAmount = parse(uint256, x, 'sourceAmount');
            self.targetAmount = parse(uint256, x, 'targetAmount');
            self.limit = parse(uint256, x, 'limit');
            self.tradingFeeAmount = parse(uint256, x, 'tradingFeeAmount');
            self.networkFeeAmount = parse(uint256, x, 'networkFeeAmount');
            self.sourceBalance = parse(uint256, x, 'sourceBalance');
            self.targetBalance = parse(uint256, x, 'targetBalance');
            self.stakedBalance = parse(uint256, x, 'stakedBalance');
            self.pool = parse(address, x, 'pool');
            self.isSourceBNT = parse(bool, x, 'isSourceBNT');
            self.bySourceAmount = parse(bool, x, 'bySourceAmount');
            self.tradingFeePPM = parse(uint32, x, 'stakedBalance');
            self.contextId = parse(uint256, x, 'sourceAmount');

    class TradeAmountAndTradingFee:
        def __init__(self, x = None) -> None:
            self.amount = parse(uint256, x, 'amount');
            self.tradingFeeAmount = parse(uint256, x, 'tradingFeeAmount');

    '''
     * @dev initializes a new PoolCollection contract
    '''
    def __init__(self,
        initNetwork,
        initBNT,
        initNetworkSettings,
        initMasterVault,
        initBNTPool,
        initExternalProtectionVault,
        initPoolTokenFactory,
        initPoolMigrator
    ) -> None:
        contract.__init__(self)
        BlockNumber.__init__(self)

        self._network = initNetwork;
        self._bnt = initBNT;
        self._networkSettings = initNetworkSettings;
        self._masterVault = initMasterVault;
        self._bntPool = initBNTPool;
        self._externalProtectionVault = initExternalProtectionVault;
        self._poolTokenFactory = initPoolTokenFactory;
        self._poolMigrator = initPoolMigrator;

        # a mapping between tokens and their pools
        self._poolData = mapping(lambda: Pool());

        # the set of all pools which are managed by this pool collection
        self._pools = EnumerableSet();

        # the default trading fee (in units of PPM)
        self._defaultTradingFeePPM = uint32();

        # the global network fee (in units of PPM)
        self._networkFeePPM = uint32();

        # true if protection is enabled, false otherwise
        self._protectionEnabled = True;

        self._setDefaultTradingFeePPM(self.DEFAULT_TRADING_FEE_PPM);
        self._setNetworkFeePPM(self.DEFAULT_NETWORK_FEE_PPM);

    '''
     * @inheritdoc IVersioned
    '''
    def version(self) -> (int):
        return 10;

    '''
     * @inheritdoc IPoolCollection
    '''
    def poolType(self) -> (uint):
        return self.POOL_TYPE.clone();

    '''
     * @inheritdoc IPoolCollection
    '''
    def defaultTradingFeePPM(self) -> (uint):
        return self._defaultTradingFeePPM.clone();

    '''
     * @inheritdoc IPoolCollection
    '''
    def networkFeePPM(self) -> (uint):
        return self._networkFeePPM.clone();

    '''
     * @inheritdoc IPoolCollection
    '''
    def pools(self) -> (list):
        return self._pools.values();

    '''
     * @inheritdoc IPoolCollection
    '''
    def poolCount(self) -> (int):
        return self._pools.length();

    '''
     * @dev sets the default trading fee (in units of PPM)
     *
     * requirements:
     *
     * - the caller must be the owner of the contract
    '''
    def setDefaultTradingFeePPM(self, newDefaultTradingFeePPM) -> None:
        self._setDefaultTradingFeePPM(newDefaultTradingFeePPM);

    '''
     * @dev sets the network fee (in units of PPM)
     *
     * requirements:
     *
     * - the caller must be the owner of the contract
    '''
    def setNetworkFeePPM(self, newNetworkFeePPM) -> None:
        self._setNetworkFeePPM(newNetworkFeePPM);

    '''
     * @dev enables/disables protection
     *
     * requirements:
     *
     * - the caller must be the owner of the contract
    '''
    def enableProtection(self, status) -> None:
        if (self._protectionEnabled == status):
            return;

        self._protectionEnabled = status;

    '''
     * @dev returns the status of the protection
    '''
    def protectionEnabled(self) -> (bool):
        return self._protectionEnabled;

    '''
     * @inheritdoc IPoolCollection
    '''
    def createPool(self, token):
        if (not self._networkSettings.isTokenWhitelisted(token)):
            revert("NotWhitelisted");

        newPoolToken = self._poolTokenFactory.createPoolToken(token);

        newPoolToken.acceptOwnership();

        newPool = Pool({
            'poolToken': newPoolToken,
            'tradingFeePPM': self._defaultTradingFeePPM,
            'tradingEnabled': False,
            'depositingEnabled': True,
            'averageRates': AverageRates({ 'blockNumber': 0, 'rate': zeroFraction112(), 'invRate': zeroFraction112() }),
            'liquidity': PoolLiquidity({ 'bntTradingLiquidity': 0, 'baseTokenTradingLiquidity': 0, 'stakedBalance': 0 })
        });

        self._addPool(token, newPool);

    '''
     * @inheritdoc IPoolCollection
    '''
    def isPoolValid(self, pool) -> (bool):
        return address(self._poolData[pool].poolToken) != address(0);

    '''
     * @dev returns specific pool's data
     *
     * notes:
     *
     * - there is no guarantee that this function will remain forward compatible,
     *   so relying on it should be avoided and instead, rely on specific getters
     *   from the IPoolCollection interface
    '''
    def poolData(self, pool) -> (Pool):
        return self._poolData[pool];

    '''
     * @inheritdoc IPoolCollection
    '''
    def poolLiquidity(self, pool) -> (PoolLiquidity):
        return self._poolData[pool].liquidity;

    '''
     * @inheritdoc IPoolCollection
    '''
    def poolToken(self, pool) -> (IPoolToken):
        return self._poolData[pool].poolToken;

    '''
     * @inheritdoc IPoolCollection
    '''
    def tradingFeePPM(self, pool) -> (uint):
        return self._poolData[pool].tradingFeePPM.clone();

    '''
     * @inheritdoc IPoolCollection
    '''
    def tradingEnabled(self, pool) -> (bool):
        return self._poolData[pool].tradingEnabled;

    '''
     * @inheritdoc IPoolCollection
    '''
    def depositingEnabled(self, pool) -> (bool):
        return self._poolData[pool].depositingEnabled;

    '''
     * @inheritdoc IPoolCollection
    '''
    def poolTokenToUnderlying(self, pool, poolTokenAmount) -> (uint):
        data = self._poolData[pool];

        return self._poolTokenToUnderlying(poolTokenAmount, data.poolToken.totalSupply(), data.liquidity.stakedBalance);

    '''
     * @inheritdoc IPoolCollection
    '''
    def underlyingToPoolToken(self, pool, baseTokenAmount) -> (uint):
        data = self._poolData[pool];

        return self._underlyingToPoolToken(baseTokenAmount, data.poolToken.totalSupply(), data.liquidity.stakedBalance);

    '''
     * @inheritdoc IPoolCollection
    '''
    def poolTokenAmountToBurn(self,
        pool,
        baseTokenAmountToDistribute,
        protocolPoolTokenAmount
    ) -> (uint):
        if (baseTokenAmountToDistribute == 0):
            return uint256(0);

        data = self._poolData[pool];

        poolTokenSupply = data.poolToken.totalSupply();
        val = baseTokenAmountToDistribute * poolTokenSupply;

        return \
            MathEx.mulDivF(
                val,
                poolTokenSupply,
                val + data.liquidity.stakedBalance * (poolTokenSupply - protocolPoolTokenAmount)
            );

    '''
     * @inheritdoc IPoolCollection
    '''
    def isPoolStable(self, pool) -> (bool):
        data = self._poolData[pool];

        return self._poolRateState(data) == PoolRateState.Stable;

    '''
     * @dev sets the trading fee of a given pool
     *
     * requirements:
     *
     * - the caller must be the owner of the contract
    '''
    def setTradingFeePPM(self, pool, newTradingFeePPM) -> None:
        data = self._poolStorage(pool);

        prevTradingFeePPM = data.tradingFeePPM;
        if (prevTradingFeePPM == newTradingFeePPM):
            return;

        data.tradingFeePPM = uint32(newTradingFeePPM);

    '''
     * @dev enables trading in a given pool, by providing the funding rate as two virtual balances, and updates its
     * trading liquidity
     *
     * note that the virtual balances should be derived from token prices, normalized to the smallest unit of
     * tokens. In other words, the ratio between BNT and TKN virtual balances should be the ratio between the $ value
     * of 1 wei of TKN and 1 wei of BNT, taking both of their decimals into account. For example:
     *
     * - if the price of one (10**18 wei) BNT is $X and the price of one (10**18 wei) TKN is $Y, then the virtual balances
     *   should represent a ratio of X to Y
     * - if the price of one (10**18 wei) BNT is $X and the price of one (10**6 wei) USDC is $Y, then the virtual balances
     *   should represent a ratio of X to Y*10**12
     *
     * requirements:
     *
     * - the caller must be the owner of the contract
    '''
    def enableTrading(self,
        pool,
        bntVirtualBalance,
        baseTokenVirtualBalance
    ) -> None:
        fundingRate = Fraction256({ 'n': bntVirtualBalance, 'd': baseTokenVirtualBalance });
        self._validRate(fundingRate);

        data = self._poolStorage(pool);

        if (data.tradingEnabled):
            revert("AlreadyEnabled");

        # adjust the trading liquidity based on the base token vault balance and funding limits
        contextId = uint256(0);
        minLiquidityForTrading = self._networkSettings.minLiquidityForTrading();
        self._updateTradingLiquidity(contextId, pool, data, fundingRate, minLiquidityForTrading);

        # verify that the BNT trading liquidity is equal or greater than the minimum liquidity for trading
        if (data.liquidity.bntTradingLiquidity < minLiquidityForTrading):
            revert("InsufficientLiquidity");

        fundingRate112 = fundingRate.toFraction112();
        data.averageRates = AverageRates({
            'blockNumber': self._blockNumber(),
            'rate': fundingRate112,
            'invRate': fundingRate112.inverse()
        });

        data.tradingEnabled = True;

    '''
     * @dev disables trading in a given pool
     *
     * requirements:
     *
     * - the caller must be the owner of the contract
    '''
    def disableTrading(self, pool) -> None:
        data = self._poolStorage(pool);
        self._resetTradingLiquidity(uint256(0), pool, data, data.liquidity, TRADING_STATUS_UPDATE_ADMIN);

    '''
     * @dev adjusts the trading liquidity in the given pool based on the base token
     * vault balance/funding limit
     *
     * requirements:
     *
     * - the caller must be the owner of the contract
    '''
    def updateTradingLiquidity(self, pool) -> None:
        data = self._poolStorage(pool);
        liquidity = data.liquidity;

        contextId = uint256(0);

        effectiveAverageRates = self._effectiveAverageRates(
            data.averageRates,
            Fraction256({ 'n': liquidity.bntTradingLiquidity, 'd': liquidity.baseTokenTradingLiquidity })
        );
        minLiquidityForTrading = self._networkSettings.minLiquidityForTrading();
        self._updateTradingLiquidity(
            contextId,
            pool,
            data,
            effectiveAverageRates.rate.fromFraction112(),
            minLiquidityForTrading
        );

    '''
     * @dev enables/disables depositing into a given pool
     *
     * requirements:
     *
     * - the caller must be the owner of the contract
    '''
    def enableDepositing(self, pool, status) -> None:
        data = self._poolStorage(pool);

        if (data.depositingEnabled == status):
            return;

        data.depositingEnabled = status;

    '''
     * @inheritdoc IPoolCollection
    '''
    def depositFor(self,
        contextId,
        provider,
        pool,
        baseTokenAmount
    ) -> (uint):
        data = self._poolStorage(pool);

        if (not data.depositingEnabled):
            revert("DepositingDisabled");

        # if there are no pool tokens available to support the staked balance - reset the
        # trading liquidity and the staked balance
        # in addition, get the effective average rates
        prevPoolTokenTotalSupply = data.poolToken.totalSupply();
        currentStakedBalance = data.liquidity.stakedBalance;
        if (prevPoolTokenTotalSupply == 0 and currentStakedBalance != 0):
            currentStakedBalance = uint256(0);

            prevLiquidity = data.liquidity;
            self._resetTradingLiquidity(contextId, pool, data, prevLiquidity, TRADING_STATUS_UPDATE_INVALID_STATE);
            effectiveAverageRates = AverageRates({
                'blockNumber': 0,
                'rate': zeroFraction112(),
                'invRate': zeroFraction112()
            });
        else:
            prevLiquidity = data.liquidity;
            effectiveAverageRates = self._effectiveAverageRates(
                data.averageRates,
                Fraction256({ 'n': prevLiquidity.bntTradingLiquidity, 'd': prevLiquidity.baseTokenTradingLiquidity })
            );

        # calculate the pool token amount to mint
        poolTokenAmount = self._underlyingToPoolToken(
            baseTokenAmount,
            prevPoolTokenTotalSupply,
            currentStakedBalance
        );

        # update the staked balance with the full base token amount
        data.liquidity.stakedBalance = currentStakedBalance + baseTokenAmount;

        # mint pool tokens to the provider
        data.poolToken.mint(provider, poolTokenAmount);

        # adjust the trading liquidity based on the base token vault balance and funding limits
        self._updateTradingLiquidity(
            contextId,
            pool,
            data,
            effectiveAverageRates.rate.fromFraction112(),
            self._networkSettings.minLiquidityForTrading()
        );

        # if trading is enabled, then update the recent average rates
        if (data.tradingEnabled):
            liquidity = data.liquidity;
            self._updateAverageRates(
                data,
                Fraction256({ 'n': liquidity.bntTradingLiquidity, 'd': liquidity.baseTokenTradingLiquidity })
            );

        return poolTokenAmount;

    '''
     * @inheritdoc IPoolCollection
    '''
    def withdraw(self,
        contextId,
        provider,
        pool,
        poolTokenAmount,
        baseTokenAmount
    ) -> (uint):
        data = self._poolStorage(pool);
        liquidity = data.liquidity;

        poolTokenTotalSupply = data.poolToken.totalSupply();
        underlyingAmount = self._poolTokenToUnderlying(
            poolTokenAmount,
            poolTokenTotalSupply,
            liquidity.stakedBalance
        );

        if (baseTokenAmount > underlyingAmount):
            revert("InvalidParam");

        if (self._poolRateState(data) == PoolRateState.Unstable):
            revert("RateUnstable");

        # obtain the withdrawal amounts
        amounts = self._poolWithdrawalAmounts(
            pool,
            poolTokenAmount,
            baseTokenAmount,
            liquidity,
            data.tradingFeePPM,
            poolTokenTotalSupply
        );

        # execute the actual withdrawal
        self._executeWithdrawal(contextId, provider, pool, data, amounts);

        # if trading is enabled, then update the recent average rates
        if (data.tradingEnabled):
            self._updateAverageRates(
                data,
                Fraction256({ 'n': data.liquidity.bntTradingLiquidity, 'd': data.liquidity.baseTokenTradingLiquidity })
            );

        return amounts.baseTokensToTransferFromMasterVault;

    '''
     * @inheritdoc IPoolCollection
    '''
    def withdrawalAmounts(self, pool, poolTokenAmount) -> (WithdrawalAmounts):
        data = self._poolData[pool];
        liquidity = data.liquidity;

        poolTokenTotalSupply = data.poolToken.totalSupply();
        underlyingAmount = self._poolTokenToUnderlying(
            poolTokenAmount,
            poolTokenTotalSupply,
            liquidity.stakedBalance
        );

        amounts = self._poolWithdrawalAmounts(
            pool,
            poolTokenAmount,
            underlyingAmount,
            liquidity,
            data.tradingFeePPM,
            poolTokenTotalSupply
        );

        return \
            WithdrawalAmounts({
                'totalAmount': amounts.baseTokensWithdrawalAmount - amounts.baseTokensWithdrawalFee,
                'baseTokenAmount': amounts.baseTokensToTransferFromMasterVault + amounts.baseTokensToTransferFromEPV,
                'bntAmount': amounts.bntToMintForProvider if self._protectionEnabled else 0
            });

    '''
     * @inheritdoc IPoolCollection
    '''
    def tradeBySourceAmount(self,
        contextId,
        sourceToken,
        targetToken,
        sourceAmount,
        minReturnAmount
    ) -> (TradeAmountAndFee):
        result = self._initTrade(
            contextId,
            sourceToken,
            targetToken,
            sourceAmount,
            minReturnAmount,
            True
        );

        self._performTrade(result);

        return \
            TradeAmountAndFee({
                'amount': result.targetAmount,
                'tradingFeeAmount': result.tradingFeeAmount,
                'networkFeeAmount': result.networkFeeAmount
            });

    '''
     * @inheritdoc IPoolCollection
    '''
    def tradeByTargetAmount(self,
        contextId,
        sourceToken,
        targetToken,
        targetAmount,
        maxSourceAmount
    ) -> (TradeAmountAndFee):
        result = self._initTrade(
            contextId,
            sourceToken,
            targetToken,
            targetAmount,
            maxSourceAmount,
            False
        );

        self._performTrade(result);

        return \
            TradeAmountAndFee({
                'amount': result.sourceAmount,
                'tradingFeeAmount': result.tradingFeeAmount,
                'networkFeeAmount': result.networkFeeAmount
            });

    '''
     * @inheritdoc IPoolCollection
    '''
    def tradeOutputAndFeeBySourceAmount(self,
        sourceToken,
        targetToken,
        sourceAmount
    ) -> (TradeAmountAndFee):
        result = self._initTrade(uint256(0), sourceToken, targetToken, sourceAmount, 1, True);

        self._processTrade(result);

        return \
            TradeAmountAndFee({
                'amount': result.targetAmount,
                'tradingFeeAmount': result.tradingFeeAmount,
                'networkFeeAmount': result.networkFeeAmount
            });

    '''
     * @inheritdoc IPoolCollection
    '''
    def tradeInputAndFeeByTargetAmount(self,
        sourceToken,
        targetToken,
        targetAmount
    ) -> (TradeAmountAndFee):
        result = self._initTrade(
            uint256(0),
            sourceToken,
            targetToken,
            targetAmount,
            type(uint256).max,
            False
        );

        self._processTrade(result);

        return \
            TradeAmountAndFee({
                'amount': result.sourceAmount,
                'tradingFeeAmount': result.tradingFeeAmount,
                'networkFeeAmount': result.networkFeeAmount
            });

    '''
     * @inheritdoc IPoolCollection
    '''
    def onFeesCollected(self, pool, feeAmount) -> None:
        if (feeAmount == 0):
            return;

        data = self._poolStorage(pool);

        # increase the staked balance by the given amount
        data.liquidity.stakedBalance += feeAmount;

    '''
     * @inheritdoc IPoolCollection
    '''
    def migratePoolIn(self, pool, data) -> None:
        self._addPool(pool, data);

        data.poolToken.acceptOwnership();

    '''
     * @inheritdoc IPoolCollection
    '''
    def migratePoolOut(self, pool, targetPoolCollection) -> None:
        cachedPoolToken = self._poolData[pool].poolToken;

        self._removePool(pool);

        cachedPoolToken.transferOwnership(address(targetPoolCollection));

    '''
     * @dev adds a pool
    '''
    def _addPool(self, pool, data) -> None:
        if (not self._pools.add(address(pool))):
            revert("AlreadyExists");

        self._poolData[pool] = data;

    '''
     * @dev removes a pool
    '''
    def _removePool(self, pool) -> None:
        if (not self._pools.remove(address(pool))):
            revert("DoesNotExist");

        del self._poolData[pool];

    '''
     * @dev returns withdrawal amounts
    '''
    def _poolWithdrawalAmounts(self,
        pool,
        poolTokenAmount,
        baseTokensWithdrawalAmount,
        liquidity,
        poolTradingFeePPM,
        poolTokenTotalSupply
    ) -> (InternalWithdrawalAmounts):
        # the base token trading liquidity of a given pool can never be higher than the base token balance of the vault
        # whenever the base token trading liquidity is updated, it is set to at most the base token balance of the vault
        baseTokenExcessAmount = pool.balanceOf(address(self._masterVault)) - liquidity.baseTokenTradingLiquidity;

        output = PoolCollectionWithdrawal.calculateWithdrawalAmounts(
            liquidity.bntTradingLiquidity,
            liquidity.baseTokenTradingLiquidity,
            baseTokenExcessAmount,
            liquidity.stakedBalance,
            pool.balanceOf(address(self._externalProtectionVault)),
            poolTradingFeePPM,
            self._networkSettings.withdrawalFeePPM(),
            baseTokensWithdrawalAmount
        );

        return \
            InternalWithdrawalAmounts({
                'baseTokensToTransferFromMasterVault': output.s,
                'bntToMintForProvider': output.t,
                'baseTokensToTransferFromEPV': output.u,
                'baseTokensTradingLiquidityDelta': output.r,
                'bntTradingLiquidityDelta': output.p,
                'bntProtocolHoldingsDelta': output.q,
                'baseTokensWithdrawalFee': output.v,
                'baseTokensWithdrawalAmount': baseTokensWithdrawalAmount,
                'poolTokenAmount': poolTokenAmount,
                'poolTokenTotalSupply': poolTokenTotalSupply,
                'newBaseTokenTradingLiquidity':
                    liquidity.baseTokenTradingLiquidity - output.r.value if output.r.isNeg else
                    liquidity.baseTokenTradingLiquidity + output.r.value,
                'newBNTTradingLiquidity':
                    liquidity.bntTradingLiquidity - output.p.value if output.p.isNeg else
                    liquidity.bntTradingLiquidity + output.p.value
            });

    '''
     * @dev executes the following actions:
     *
     * - burn the network's base pool tokens
     * - update the pool's base token staked balance
     * - update the pool's base token trading liquidity
     * - update the pool's BNT trading liquidity
     * - update the pool's trading liquidity product
     * - emit an event if the pool's BNT trading liquidity has crossed the minimum threshold
     *   (either above the threshold or below the threshold)
    '''
    def _executeWithdrawal(self,
        contextId,
        provider,
        pool,
        data,
        amounts
    ) -> None:
        liquidity = data.liquidity;
        prevLiquidity = liquidity;

        data.poolToken.burn(amounts.poolTokenAmount);

        newPoolTokenTotalSupply = amounts.poolTokenTotalSupply - amounts.poolTokenAmount;
        newStakedBalance = MathEx.mulDivF(
            liquidity.stakedBalance,
            newPoolTokenTotalSupply,
            amounts.poolTokenTotalSupply
        );

        liquidity.stakedBalance = newStakedBalance;

        # trading liquidity is assumed to never exceed 128 bits (the cast below will revert("otherwise)
        liquidity.baseTokenTradingLiquidity = amounts.newBaseTokenTradingLiquidity.toUint128();
        liquidity.bntTradingLiquidity = amounts.newBNTTradingLiquidity.toUint128();

        if (amounts.bntProtocolHoldingsDelta.value > 0):
            assert(amounts.bntProtocolHoldingsDelta.isNeg); # currently no support for requesting funding here

            self._bntPool.renounceFunding(contextId, pool, amounts.bntProtocolHoldingsDelta.value);
        elif (amounts.bntTradingLiquidityDelta.value > 0):
            if (amounts.bntTradingLiquidityDelta.isNeg):
                self._bntPool.burnFromVault(amounts.bntTradingLiquidityDelta.value);
            else:
                self._bntPool.mint(address(self._masterVault), amounts.bntTradingLiquidityDelta.value);

        # if the provider should receive some BNT - ask the BNT pool to mint BNT to the provider
        isProtectionEnabled = self._protectionEnabled;
        if (amounts.bntToMintForProvider > 0 and isProtectionEnabled):
            self._bntPool.mint(address(provider), amounts.bntToMintForProvider);

        # if the provider should receive some base tokens from the external protection vault - remove the tokens from
        # the external protection vault and send them to the master vault
        if (amounts.baseTokensToTransferFromEPV > 0):
            self._externalProtectionVault.withdrawFunds(
                pool,
                payable(address(self._masterVault)),
                amounts.baseTokensToTransferFromEPV
            );
            amounts.baseTokensToTransferFromMasterVault += amounts.baseTokensToTransferFromEPV;

        # if the provider should receive some base tokens from the master vault - remove the tokens from the master
        # vault and send them to the provider
        if (amounts.baseTokensToTransferFromMasterVault > 0):
            self._masterVault.withdrawFunds(pool, payable(provider), amounts.baseTokensToTransferFromMasterVault);

        # ensure that the average rate is reset when the pool is being emptied
        if (amounts.newBaseTokenTradingLiquidity == 0):
            data.averageRates.rate = zeroFraction112();
            data.averageRates.invRate = zeroFraction112();

        # if the new BNT trading liquidity is below the minimum liquidity for trading - reset the liquidity
        if (amounts.newBNTTradingLiquidity < self._networkSettings.minLiquidityForTrading()):
            self._resetTradingLiquidity_(
                contextId,
                pool,
                data,
                prevLiquidity,
                amounts.newBNTTradingLiquidity,
                TRADING_STATUS_UPDATE_MIN_LIQUIDITY
            );
        else:
            self._dispatchTradingLiquidityEvents(contextId, pool, prevLiquidity, liquidity);

    '''
     * @dev sets the default trading fee (in units of PPM)
    '''
    def _setDefaultTradingFeePPM(self, newDefaultTradingFeePPM) -> None:
        prevDefaultTradingFeePPM = self._defaultTradingFeePPM;
        if (prevDefaultTradingFeePPM == newDefaultTradingFeePPM):
            return;

        self._defaultTradingFeePPM = uint32(newDefaultTradingFeePPM);

    '''
     * @dev sets the network fee (in units of PPM)
    '''
    def _setNetworkFeePPM(self, newNetworkFeePPM) -> None:
        prevNetworkFeePPM = self._networkFeePPM;
        if (prevNetworkFeePPM == newNetworkFeePPM):
            return;

        self._networkFeePPM = uint32(newNetworkFeePPM);

    '''
     * @dev returns a storage reference to pool data
    '''
    def _poolStorage(self, pool) -> (Pool):
        data = self._poolData[pool];
        if (address(data.poolToken) == address(0)):
            revert("DoesNotExist");

        return data;

    '''
     * @dev calculates base tokens amount
    '''
    def _poolTokenToUnderlying(self,
        poolTokenAmount,
        poolTokenSupply,
        stakedBalance
    ) -> (uint):
        if (poolTokenSupply == 0):
            # if this is the initial liquidity provision - use a one-to-one pool token to base token rate
            if (stakedBalance > 0):
                revert("InvalidStakedBalance");

            return poolTokenAmount;

        return MathEx.mulDivF(poolTokenAmount, stakedBalance, poolTokenSupply);

    '''
     * @dev calculates pool tokens amount
    '''
    def _underlyingToPoolToken(self,
        baseTokenAmount,
        poolTokenSupply,
        stakedBalance
    ) -> (uint):
        if (poolTokenSupply == 0):
            # if this is the initial liquidity provision - use a one-to-one pool token to base token rate
            if (stakedBalance > 0):
                revert("InvalidStakedBalance");

            return baseTokenAmount;

        return MathEx.mulDivC(baseTokenAmount, poolTokenSupply, stakedBalance);

    '''
     * @dev calculates the target trading liquidities, taking into account the total out-of-curve base token liquidity,
     * and the deltas between the new and the previous states
    '''
    def _calcTargetTradingLiquidity(self,
        baseTokenTotalLiquidity,
        fundingLimit,
        currentFunding,
        liquidity,
        fundingRate,
        minLiquidityForTrading
    ) -> (TargetTradingLiquidity):
        # calculate the target BNT trading liquidity based on the following:
        # - BNT liquidity required to match the based token unused (off-curve) liquidity
        # - BNT funding limit
        # - current BNT funding
        targetBNTTradingLiquidity = liquidity.bntTradingLiquidity;
        if (fundingLimit > currentFunding):
            # increase the trading liquidity
            availableFunding = fundingLimit - currentFunding;
            baseTokenUnusedLiquidity = baseTokenTotalLiquidity - liquidity.baseTokenTradingLiquidity;
            targetBNTTradingLiquidityDelta = Math.min(
                MathEx.mulDivF(baseTokenUnusedLiquidity, fundingRate.n, fundingRate.d),
                availableFunding
            );
            targetBNTTradingLiquidity = liquidity.bntTradingLiquidity + targetBNTTradingLiquidityDelta;
        elif (fundingLimit < currentFunding):
            # decrease the trading liquidity
            excessFunding = currentFunding - fundingLimit;
            targetBNTTradingLiquidity = MathEx.subMax0(liquidity.bntTradingLiquidity, excessFunding);

        # if the target is equal to the current trading liquidity, no update is needed
        if (targetBNTTradingLiquidity == liquidity.bntTradingLiquidity):
            return TargetTradingLiquidity({ 'update': False, 'bnt': 0, 'baseToken': 0 });

        # ensure that the target is above the minimum liquidity for trading
        if (targetBNTTradingLiquidity < minLiquidityForTrading):
            return TargetTradingLiquidity({ 'update': True, 'bnt': 0, 'baseToken': 0 });

        # calculate the target base token trading liquidity using the following:
        # - calculate the delta between the current/target BNT trading liquidity
        # - calculate the base token trading liquidity delta based on the BNT trading liquidity delta and the funding rate
        # - apply the base token trading liquidity delta to the current base token trading liquidity
        #
        # note that the effective funding rate is always the rate between BNT and the base token)

        # liquidity increase
        # note that liquidity increase is capped
        if (targetBNTTradingLiquidity > liquidity.bntTradingLiquidity):
            if (liquidity.bntTradingLiquidity == 0):
                # the current BNT trading liquidity is 0 - cap the target trading liquidity
                # by the default bootstrap amount, which includes a buffer to reduce the chance
                # for trading to be disabled as a result of trades in the pool
                tradingLiquidityCap = minLiquidityForTrading * self.BOOTSTRAPPING_LIQUIDITY_BUFFER_FACTOR;
            else:
                # the current BNT trading liquidity is not 0 - cap the target using the growth factor
                tradingLiquidityCap = liquidity.bntTradingLiquidity * self.LIQUIDITY_GROWTH_FACTOR;

            # apply the trading liquidity cap
            targetBNTTradingLiquidity = Math.min(targetBNTTradingLiquidity, tradingLiquidityCap);

            # calculate the trading liquidity deltas and return them
            bntTradingLiquidityDelta = targetBNTTradingLiquidity - liquidity.bntTradingLiquidity;
            baseTokenTradingLiquidityDelta = MathEx.mulDivF(bntTradingLiquidityDelta, fundingRate.d, fundingRate.n);

            return \
                TargetTradingLiquidity({
                    'update': True,
                    'bnt': targetBNTTradingLiquidity,
                    'baseToken': liquidity.baseTokenTradingLiquidity + baseTokenTradingLiquidityDelta
                });

        # liquidity decrease
        # note that liquidity decrease isn't capped
        # calculate the trading liquidity deltas and return them
        bntTradingLiquidityDelta = liquidity.bntTradingLiquidity - targetBNTTradingLiquidity;
        baseTokenTradingLiquidityDelta = MathEx.mulDivF(bntTradingLiquidityDelta, fundingRate.d, fundingRate.n);

        return \
            TargetTradingLiquidity({
                'update': True,
                'bnt': targetBNTTradingLiquidity,
                'baseToken': MathEx.subMax0(liquidity.baseTokenTradingLiquidity, baseTokenTradingLiquidityDelta)
            });

    '''
     * @dev adjusts the trading liquidity based on the newly added tokens delta amount, and funding limits
    '''
    def _updateTradingLiquidity(self,
        contextId,
        pool,
        data,
        fundingRate,
        minLiquidityForTrading
    ) -> None:
        # ensure that the base token reserve isn't empty
        baseTokenTotalLiquidity = pool.balanceOf(address(self._masterVault));
        if (baseTokenTotalLiquidity == 0):
            revert("InsufficientLiquidity");

        if (self._poolRateState(data) == PoolRateState.Unstable):
            return;

        prevLiquidity = data.liquidity;
        if (not fundingRate.isPositive()):
            self._resetTradingLiquidity(contextId, pool, data, prevLiquidity, TRADING_STATUS_UPDATE_MIN_LIQUIDITY);

            return;

        targetTradingLiquidity = self._calcTargetTradingLiquidity(
            baseTokenTotalLiquidity,
            self._networkSettings.poolFundingLimit(pool),
            self._bntPool.currentPoolFunding(pool),
            prevLiquidity,
            fundingRate,
            minLiquidityForTrading
        );

        if (not targetTradingLiquidity.update):
            return;

        if (targetTradingLiquidity.bnt == 0 or targetTradingLiquidity.baseToken == 0):
            self._resetTradingLiquidity(contextId, pool, data, prevLiquidity, TRADING_STATUS_UPDATE_MIN_LIQUIDITY);
            return;

        # update funding from the BNT pool
        if (targetTradingLiquidity.bnt > prevLiquidity.bntTradingLiquidity):
            self._bntPool.requestFunding(contextId, pool, targetTradingLiquidity.bnt - prevLiquidity.bntTradingLiquidity);
        elif (targetTradingLiquidity.bnt < prevLiquidity.bntTradingLiquidity):
            self._bntPool.renounceFunding(contextId, pool, prevLiquidity.bntTradingLiquidity - targetTradingLiquidity.bnt);

        # trading liquidity is assumed to never exceed 128 bits (the cast below will revert("otherwise)
        newLiquidity = PoolLiquidity({
            'bntTradingLiquidity': targetTradingLiquidity.bnt.toUint128(),
            'baseTokenTradingLiquidity': targetTradingLiquidity.baseToken.toUint128(),
            'stakedBalance': prevLiquidity.stakedBalance
        });

        # update the liquidity data of the pool
        data.liquidity = newLiquidity;

        self._dispatchTradingLiquidityEvents(contextId, pool, prevLiquidity, newLiquidity);

    def _dispatchTradingLiquidityEvents(self,
        contextId,
        pool,
        prevLiquidity,
        newLiquidity
    ) -> None:
        if (newLiquidity.bntTradingLiquidity != prevLiquidity.bntTradingLiquidity):
            pass

        if (newLiquidity.baseTokenTradingLiquidity != prevLiquidity.baseTokenTradingLiquidity):
            pass

    '''
     * @dev resets trading liquidity and renounces any remaining BNT funding
    '''
    def _resetTradingLiquidity(self,
        contextId,
        pool,
        data,
        prevLiquidity,
        reason
    ) -> None:
        self._resetTradingLiquidity_(contextId, pool, data, prevLiquidity, data.liquidity.bntTradingLiquidity, reason);

    '''
     * @dev resets trading liquidity and renounces any remaining BNT funding
    '''
    def _resetTradingLiquidity_(self,
        contextId,
        pool,
        data,
        prevLiquidity,
        currentBNTTradingLiquidity,
        reason
    ) -> None:
        # reset the network and base token trading liquidities
        data.liquidity.bntTradingLiquidity = uint128(0);
        data.liquidity.baseTokenTradingLiquidity = uint128(0);

        # reset the recent average rage
        data.averageRates = AverageRates({ 'blockNumber': 0, 'rate': zeroFraction112(), 'invRate': zeroFraction112() });

        # ensure that trading is disabled
        if (data.tradingEnabled):
            data.tradingEnabled = False;

        # renounce all network liquidity
        if (currentBNTTradingLiquidity > 0):
            self._bntPool.renounceFunding(contextId, pool, currentBNTTradingLiquidity);

        self._dispatchTradingLiquidityEvents(contextId, pool, prevLiquidity, data.liquidity);

    '''
     * @dev returns initial trading params
    '''
    def _initTrade(self,
        contextId,
        sourceToken,
        targetToken,
        amount,
        limit,
        bySourceAmount
    ) -> (TradeIntermediateResult):
        result = self.TradeIntermediateResult()

        # ensure that BNT is either the source or the target token
        isSourceBNT = sourceToken is (self._bnt);
        isTargetBNT = targetToken is (self._bnt);

        if (isSourceBNT and not isTargetBNT):
            result.isSourceBNT = True;
            result.pool = targetToken;
        elif (not isSourceBNT and isTargetBNT):
            result.isSourceBNT = False;
            result.pool = sourceToken;
        else:
            # BNT isn't one of the tokens or is both of them
            revert("DoesNotExist");

        data = self._poolStorage(result.pool);

        # verify that trading is enabled
        if (not data.tradingEnabled):
            revert("TradingDisabled");

        result.contextId = contextId;
        result.bySourceAmount = bySourceAmount;

        if (result.bySourceAmount):
            result.sourceAmount = amount;
        else:
            result.targetAmount = amount;

        result.limit = limit;
        result.tradingFeePPM = data.tradingFeePPM;

        liquidity = data.liquidity;
        if (result.isSourceBNT):
            result.sourceBalance = uint256(liquidity.bntTradingLiquidity);
            result.targetBalance = uint256(liquidity.baseTokenTradingLiquidity);
        else:
            result.sourceBalance = uint256(liquidity.baseTokenTradingLiquidity);
            result.targetBalance = uint256(liquidity.bntTradingLiquidity);

        result.stakedBalance = liquidity.stakedBalance;

        return result

    '''
     * @dev returns trade amount and fee by providing the source amount
    '''
    def _tradeAmountAndFeeBySourceAmount(self,
        sourceBalance,
        targetBalance,
        feePPM,
        sourceAmount
    ) -> (TradeAmountAndTradingFee):
        if (sourceBalance == 0 or targetBalance == 0):
            revert("InsufficientLiquidity");

        targetAmount = MathEx.mulDivF(targetBalance, sourceAmount, sourceBalance + sourceAmount);
        tradingFeeAmount = MathEx.mulDivF(targetAmount, feePPM, PPM_RESOLUTION);

        return \
            self.TradeAmountAndTradingFee({ 'amount': targetAmount - tradingFeeAmount, 'tradingFeeAmount': tradingFeeAmount });

    '''
     * @dev returns trade amount and fee by providing either the target amount
    '''
    def _tradeAmountAndFeeByTargetAmount(self,
        sourceBalance,
        targetBalance,
        feePPM,
        targetAmount
    ) -> (TradeAmountAndTradingFee):
        if (sourceBalance == 0):
            revert("InsufficientLiquidity");

        tradingFeeAmount = MathEx.mulDivF(targetAmount, feePPM, PPM_RESOLUTION - feePPM);
        fullTargetAmount = targetAmount + tradingFeeAmount;
        sourceAmount = MathEx.mulDivF(sourceBalance, fullTargetAmount, targetBalance - fullTargetAmount);

        return self.TradeAmountAndTradingFee({ 'amount': sourceAmount, 'tradingFeeAmount': tradingFeeAmount });

    '''
     * @dev processes a trade by providing either the source or the target amount and updates the in-memory intermediate
     * result
    '''
    def _processTrade(self, result) -> None:
        tradeAmountAndFee = self.TradeAmountAndTradingFee();

        if (result.bySourceAmount):
            tradeAmountAndFee = self._tradeAmountAndFeeBySourceAmount(
                result.sourceBalance,
                result.targetBalance,
                result.tradingFeePPM,
                result.sourceAmount
            );

            result.targetAmount = tradeAmountAndFee.amount;

            # ensure that the target amount is above the requested minimum return amount
            if (result.targetAmount < result.limit):
                revert("InsufficientTargetAmount");
        else:
            tradeAmountAndFee = self._tradeAmountAndFeeByTargetAmount(
                result.sourceBalance,
                result.targetBalance,
                result.tradingFeePPM,
                result.targetAmount
            );

            result.sourceAmount = tradeAmountAndFee.amount;

            # ensure that the user has provided enough tokens to make the trade
            if (result.sourceAmount == 0 or result.sourceAmount > result.limit):
                revert("InsufficientSourceAmount");

        result.tradingFeeAmount = tradeAmountAndFee.tradingFeeAmount;

        # sync the trading and staked balance
        result.sourceBalance += result.sourceAmount;
        result.targetBalance -= result.targetAmount;

        if (result.isSourceBNT):
            result.stakedBalance += result.tradingFeeAmount;

        self._processNetworkFee(result);

    '''
     * @dev processes the network fee and updates the in-memory intermediate result
    '''
    def _processNetworkFee(self, result) -> None:
        if (self._networkFeePPM == 0):
            return;

        # calculate the target network fee amount
        targetNetworkFeeAmount = MathEx.mulDivF(result.tradingFeeAmount, self._networkFeePPM, PPM_RESOLUTION);

        # update the target balance (but don't deduct it from the full trading fee amount)
        result.targetBalance -= targetNetworkFeeAmount;

        if (not result.isSourceBNT):
            result.networkFeeAmount = targetNetworkFeeAmount;

            return;

        # trade the network fee (taken from the base token) to BNT
        result.networkFeeAmount = self._tradeAmountAndFeeBySourceAmount(
            result.targetBalance,
            result.sourceBalance,
            0,
            targetNetworkFeeAmount
        ).amount;

        # since we have received the network fee in base tokens and have traded them for BNT (so that the network fee
        # is always kept in BNT), we'd need to adapt the trading liquidity and the staked balance accordingly
        result.targetBalance += targetNetworkFeeAmount;
        result.sourceBalance -= result.networkFeeAmount;
        result.stakedBalance -= targetNetworkFeeAmount;

    '''
     * @dev performs a trade
    '''
    def _performTrade(self, result) -> None:
        data = self._poolData[result.pool];
        prevLiquidity = data.liquidity;

        # update the recent average rate
        self._updateAverageRates(
            data,
            Fraction256({ 'n': prevLiquidity.bntTradingLiquidity, 'd': prevLiquidity.baseTokenTradingLiquidity })
        );

        self._processTrade(result);

        # trading liquidity is assumed to never exceed 128 bits (the cast below will revert("otherwise)
        newLiquidity = PoolLiquidity({
            'bntTradingLiquidity': (result.sourceBalance if result.isSourceBNT else result.targetBalance).toUint128(),
            'baseTokenTradingLiquidity': (result.targetBalance if result.isSourceBNT else result.sourceBalance).toUint128(),
            'stakedBalance': result.stakedBalance
        });

        self._dispatchTradingLiquidityEvents(result.contextId, result.pool, prevLiquidity, newLiquidity);

        # update the liquidity data of the pool
        data.liquidity = newLiquidity;

    '''
     * @dev returns the state of a pool's rate
    '''
    def _poolRateState(self, data) -> (PoolRateState):
        spotRate = Fraction256({
            'n': data.liquidity.bntTradingLiquidity,
            'd': data.liquidity.baseTokenTradingLiquidity
        });

        averageRates = data.averageRates;
        rate = averageRates.rate;
        if (not spotRate.isPositive() or not rate.isPositive()):
            return PoolRateState.Uninitialized;

        invSpotRate = spotRate.inverse();
        invRate = averageRates.invRate;
        if (not invSpotRate.isPositive() or not invRate.isPositive()):
            return PoolRateState.Uninitialized;

        effectiveAverageRates = self._effectiveAverageRates(averageRates, spotRate);

        if (
            MathEx.isInRange(effectiveAverageRates.rate.fromFraction112(), spotRate, self.RATE_MAX_DEVIATION_PPM) and
            MathEx.isInRange(effectiveAverageRates.invRate.fromFraction112(), invSpotRate, self.RATE_MAX_DEVIATION_PPM)
        ):
            return PoolRateState.Stable;

        return PoolRateState.Unstable;

    '''
     * @dev updates the average rates
    '''
    def _updateAverageRates(self, data, spotRate) -> None:
        data.averageRates = self._effectiveAverageRates(data.averageRates, spotRate);

    '''
     * @dev returns the effective average rates
    '''
    def _effectiveAverageRates(self, averageRates, spotRate) -> (AverageRates):
        # if the spot rate is 0, reset the average rates
        if (not spotRate.isPositive()):
            return AverageRates({ 'blockNumber': 0, 'rate': zeroFraction112(), 'invRate': zeroFraction112() });

        blockNumber = self._blockNumber();

        # can only be updated once in a single block
        prevUpdateBlock = averageRates.blockNumber;
        if (prevUpdateBlock == blockNumber):
            return averageRates;

        # if sufficient blocks have passed, or if one of the rates isn't positive,
        # reset the average rates
        if (
            blockNumber - prevUpdateBlock >= self.RATE_RESET_BLOCK_THRESHOLD or
            not averageRates.rate.isPositive() or
            not averageRates.invRate.isPositive()
        ):
            if (spotRate.isPositive()):
                return \
                    AverageRates({
                        'blockNumber': blockNumber,
                        'rate': spotRate.toFraction112(),
                        'invRate': spotRate.inverse().toFraction112()
                    });

            return AverageRates({ 'blockNumber': 0, 'rate': zeroFraction112(), 'invRate': zeroFraction112() });

        return \
            AverageRates({
                'blockNumber': blockNumber,
                'rate': self._calcAverageRate(averageRates.rate, spotRate),
                'invRate': self._calcAverageRate(averageRates.invRate, spotRate.inverse())
            });

    '''
     * @dev calculates the average rate
    '''
    def _calcAverageRate(self, averageRate, rate) -> (Fraction112):
        if (rate.n * averageRate.d == rate.d * averageRate.n):
            return averageRate;

        return \
            MathEx \
                .weightedAverage(averageRate.fromFraction112(), rate, self.EMA_AVERAGE_RATE_WEIGHT, self.EMA_SPOT_RATE_WEIGHT) \
                .toFraction112();

    '''
     * @dev verifies if the provided rate is valid
    '''
    def _validRate(self, rate) -> None:
        if (not rate.isPositive()):
            revert("InvalidRate");



from sys import modules

full_precision_mode = False

def enable_full_precision_mode(state):
    global full_precision_mode
    if full_precision_mode != state:
        full_precision_mode = state
        par_name = '.'.join(__name__.split('.')[:-1])
        for name in [name for name in modules if name.startswith(par_name) and name != __name__]:
            del modules[name]
